"""Logging utilities."""
import logging
import numbers
import os
import sys
import threading
import traceback
from contextlib import contextmanager
from typing import AnyStr, Sequence  # noqa

from kombu.log import LOG_LEVELS
from kombu.log import get_logger as _get_logger
from kombu.utils.encoding import safe_str

from .term import colored

__all__ = (
    'ColorFormatter', 'LoggingProxy', 'base_logger',
    'set_in_sighandler', 'in_sighandler', 'get_logger',
    'get_task_logger', 'mlevel',
    'get_multiprocessing_logger', 'reset_multiprocessing_logger', 'LOG_LEVELS'
)

_process_aware = False
_in_sighandler = False

MP_LOG = os.environ.get('MP_LOG', False)

RESERVED_LOGGER_NAMES = {'celery', 'celery.task'}

# Sets up our logging hierarchy.
#
# Every logger in the celery package inherits from the "celery"
# logger, and every task logger inherits from the "celery.task"
# logger.
base_logger = logger = _get_logger('celery')


def set_in_sighandler(value):
    """Set flag signifying that we're inside a signal handler."""
    global _in_sighandler
    _in_sighandler = value


def iter_open_logger_fds():
    seen = set()
    loggers = (list(logging.Logger.manager.loggerDict.values()) +
               [logging.getLogger(None)])
    for l in loggers:
        try:
            for handler in l.handlers:
                try:
                    if handler not in seen:  # pragma: no cover
                        yield handler.stream
                        seen.add(handler)
                except AttributeError:
                    pass
        except AttributeError:  # PlaceHolder does not have handlers
            pass


@contextmanager
def in_sighandler():
    """Context that records that we are in a signal handler."""
    set_in_sighandler(True)
    try:
        yield
    finally:
        set_in_sighandler(False)


def logger_isa(l, p, max=1000):
    this, seen = l, set()
    for _ in range(max):
        if this == p:
            return True
        else:
            if this in seen:
                raise RuntimeError(
                    f'Logger {l.name!r} parents recursive',
                )
            seen.add(this)
            this = this.parent
            if not this:
                break
    else:  # pragma: no cover
        raise RuntimeError(f'Logger hierarchy exceeds {max}')
    return False


def _using_logger_parent(parent_logger, logger_):
    if not logger_isa(logger_, parent_logger):
        logger_.parent = parent_logger
    return logger_


def get_logger(name):
    """Get logger by name."""
    l = _get_logger(name)
    if logging.root not in (l, l.parent) and l is not base_logger:
        l = _using_logger_parent(base_logger, l)
    return l


task_logger = get_logger('celery.task')
worker_logger = get_logger('celery.worker')


def get_task_logger(name):
    """Get logger for task module by name."""
    if name in RESERVED_LOGGER_NAMES:
        raise RuntimeError(f'Logger name {name!r} is reserved!')
    return _using_logger_parent(task_logger, get_logger(name))


def mlevel(level):
    """Convert level name/int to log level."""
    if level and not isinstance(level, numbers.Integral):
        return LOG_LEVELS[level.upper()]
    return level


class ColorFormatter(logging.Formatter):
    """Logging formatter that adds colors based on severity."""

    #: Loglevel -> Color mapping.
    COLORS = colored().names
    colors = {
        'DEBUG': COLORS['blue'],
        'WARNING': COLORS['yellow'],
        'ERROR': COLORS['red'],
        'CRITICAL': COLORS['magenta'],
    }

    def __init__(self, fmt=None, use_color=True):
        super().__init__(fmt)
        self.use_color = use_color

    def formatException(self, ei):
        if ei and not isinstance(ei, tuple):
            ei = sys.exc_info()
        r = super().formatException(ei)
        return r

    def format(self, record):
        msg = super().format(record)
        color = self.colors.get(record.levelname)

        # reset exception info later for other handlers...
        einfo = sys.exc_info() if record.exc_info == 1 else record.exc_info

        if color and self.use_color:
            try:
                # safe_str will repr the color object
                # and color will break on non-string objects
                # so need to reorder calls based on type.
                # Issue #427
                try:
                    if isinstance(msg, str):
                        return str(color(safe_str(msg)))
                    return safe_str(color(msg))
                except UnicodeDecodeError:  # pragma: no cover
                    return safe_str(msg)  # skip colors
            except Exception as exc:  # pylint: disable=broad-except
                prev_msg, record.exc_info, record.msg = (
                    record.msg, 1, '<Unrepresentable {!r}: {!r}>'.format(
                        type(msg), exc
                    ),
                )
                try:
                    return super().format(record)
                finally:
                    record.msg, record.exc_info = prev_msg, einfo
        else:
            return safe_str(msg)


class LoggingProxy:
    """Forward file object to :class:`logging.Logger` instance.

    Arguments:
        logger (~logging.Logger): Logger instance to forward to.
        loglevel (int, str): Log level to use when logging messages.
    """

    mode = 'w'
    name = None
    closed = False
    loglevel = logging.ERROR
    _thread = threading.local()

    def __init__(self, logger, loglevel=None):
        # pylint: disable=redefined-outer-name
        # Note that the logger global is redefined here, be careful changing.
        self.logger = logger
        self.loglevel = mlevel(loglevel or self.logger.level or self.loglevel)
        self._safewrap_handlers()

    def _safewrap_handlers(self):
        # Make the logger handlers dump internal errors to
        # :data:`sys.__stderr__` instead of :data:`sys.stderr` to circumvent
        # infinite loops.

        def wrap_handler(handler):                  # pragma: no cover

            class WithSafeHandleError(logging.Handler):

                def handleError(self, record):
                    try:
                        traceback.print_exc(None, sys.__stderr__)
                    except OSError:
                        pass    # see python issue 5971

            handler.handleError = WithSafeHandleError().handleError
        return [wrap_handler(h) for h in self.logger.handlers]

    def write(self, data):
        # type: (AnyStr) -> int
        """Write message to logging object."""
        if _in_sighandler:
            safe_data = safe_str(data)
            print(safe_data, file=sys.__stderr__)
            return len(safe_data)
        if getattr(self._thread, 'recurse_protection', False):
            # Logger is logging back to this file, so stop recursing.
            return 0
        if data and not self.closed:
            self._thread.recurse_protection = True
            try:
                safe_data = safe_str(data).rstrip('\n')
                if safe_data:
                    self.logger.log(self.loglevel, safe_data)
                    return len(safe_data)
            finally:
                self._thread.recurse_protection = False
        return 0

    def writelines(self, sequence):
        # type: (Sequence[str]) -> None
        """Write list of strings to file.

        The sequence can be any iterable object producing strings.
        This is equivalent to calling :meth:`write` for each string.
        """
        for part in sequence:
            self.write(part)

    def flush(self):
        # This object is not buffered so any :meth:`flush`
        # requests are ignored.
        pass

    def close(self):
        # when the object is closed, no write requests are
        # forwarded to the logging object anymore.
        self.closed = True

    def isatty(self):
        """Here for file support."""
        return False


def get_multiprocessing_logger():
    """Return the multiprocessing logger."""
    try:
        from billiard import util
    except ImportError:
        pass
    else:
        return util.get_logger()


def reset_multiprocessing_logger():
    """Reset multiprocessing logging setup."""
    try:
        from billiard import util
    except ImportError:
        pass
    else:
        if hasattr(util, '_logger'):  # pragma: no cover
            util._logger = None


def current_process():
    try:
        from billiard import process
    except ImportError:
        pass
    else:
        return process.current_process()


def current_process_index(base=1):
    index = getattr(current_process(), 'index', None)
    return index + base if index is not None else index



from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING
from typing import Any

if TYPE_CHECKING:
    from .event import Event
    from .state import State
    from .statemachine import StateMachine
    from .transition import Transition


@dataclass
class TriggerData:
    machine: "StateMachine"

    event: "Event"
    """The Event that was triggered."""

    model: Any = field(init=False)
    """A reference to the underlying model that holds the current :ref:`State`."""

    args: tuple = field(default_factory=tuple)
    """All positional arguments provided on the :ref:`Event`."""

    kwargs: dict = field(default_factory=dict)
    """All keyword arguments provided on the :ref:`Event`."""

    def __post_init__(self):
        self.model = self.machine.model


@dataclass
class EventData:
    trigger_data: TriggerData
    """The :ref:`TriggerData` of the :ref:`event`."""

    transition: "Transition"
    """The :ref:`Transition` instance that was activated by the :ref:`Event`."""

    state: "State" = field(init=False)
    """The current :ref:`State` of the :ref:`statemachine`."""

    source: "State" = field(init=False)
    """The :ref:`State` which :ref:`statemachine` was in when the Event started."""

    target: "State" = field(init=False)
    """The destination :ref:`State` of the :ref:`transition`."""

    result: "Any | None" = None

    executed: bool = False

    def __post_init__(self):
        self.state = self.transition.source
        self.source = self.transition.source
        self.target = self.transition.target
        self.machine = self.trigger_data.machine

    @property
    def event(self):
        return self.trigger_data.event

    @property
    def args(self):
        return self.trigger_data.args

    @property
    def extended_kwargs(self):
        kwargs = self.trigger_data.kwargs.copy()
        kwargs["event_data"] = self
        kwargs["machine"] = self.trigger_data.machine
        kwargs["event"] = self.trigger_data.event
        kwargs["model"] = self.trigger_data.model
        kwargs["transition"] = self.transition
        kwargs["state"] = self.state
        kwargs["source"] = self.source
        kwargs["target"] = self.target
        return kwargs



import gettext
from pathlib import Path

script_dir = Path(__file__).resolve().parent
locale_dir = script_dir / "locale"


def setup_i18n():
    translate = gettext.translation("statemachine", locale_dir, fallback=True)
    gettext.bindtextdomain("statemachine", locale_dir)
    gettext.textdomain("statemachine")
    return translate.gettext


_ = setup_i18n()



from dataclasses import dataclass
from functools import partial
from functools import reduce
from operator import attrgetter
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Iterable
from typing import List
from typing import Set
from typing import Tuple

from .callbacks import SPECS_ALL
from .callbacks import SpecReference
from .callbacks import allways_true
from .event import Event
from .exceptions import InvalidDefinition
from .i18n import _
from .signature import SignatureAdapter
from .spec_parser import custom_and
from .spec_parser import operator_mapping
from .spec_parser import parse_boolean_expr

if TYPE_CHECKING:
    from .callbacks import CallbackSpec
    from .callbacks import CallbackSpecList
    from .callbacks import CallbacksRegistry


@dataclass
class Listener:
    """Object reference that provides attributes to be used as callbacks.

    Args:
        obj: Any object that will serve as lookup for attributes.
        skip_attrs: Protected attrs that will be ignored on the search.
    """

    obj: object
    all_attrs: Set[str]
    resolver_id: str

    @classmethod
    def from_obj(cls, obj, skip_attrs=None) -> "Listener":
        if isinstance(obj, Listener):
            return obj
        else:
            if skip_attrs is None:
                skip_attrs = set()
            all_attrs = set(dir(obj)) - skip_attrs
            return cls(obj, all_attrs, str(id(obj)))

    def build_key(self, attr_name) -> str:
        return f"{attr_name}@{self.resolver_id}"


@dataclass
class Listeners:
    """Listeners that provides attributes to be used as callbacks."""

    items: Tuple[Listener, ...]
    all_attrs: Set[str]

    @classmethod
    def from_listeners(cls, listeners: Iterable["Listener"]) -> "Listeners":
        listeners = tuple(listeners)
        all_attrs = set().union(*(listener.all_attrs for listener in listeners))
        return cls(listeners, all_attrs)

    def resolve(
        self,
        specs: "CallbackSpecList",
        registry: "CallbacksRegistry",
        allowed_references: SpecReference = SPECS_ALL,
    ):
        found_convention_specs = specs.conventional_specs & self.all_attrs

        for spec in specs:
            if (spec.reference not in allowed_references) or (
                spec.is_convention and spec.func not in found_convention_specs
            ):
                continue

            executor = registry[specs.grouper(spec.group).key]
            for key, builder in self.build(spec):
                executor.add(key, spec, builder)

    def _take_callback(self, name: str, names_not_found_handler: Callable) -> Callable:
        callbacks: List[Callable] = []
        for key, builder in self.search_name(name):
            callback = builder()
            callback.unique_key = key  # type: ignore[attr-defined]
            callbacks.append(callback)

        if len(callbacks) == 0:
            names_not_found_handler(name)
            return allways_true
        elif len(callbacks) == 1:
            return callbacks[0]
        else:
            return reduce(custom_and, callbacks)

    def build(self, spec: "CallbackSpec"):
        """
        Resolves the `spec` into callables in the `registry`.

        Args:
            spec (CallbackSpec): A spec to be resolved.
            registry (callable): A callable that will be used to store the resolved callables.
        """
        if not spec.may_contain_boolean_expression:
            yield from self.search(spec)
            return

        # Resolves boolean expressions

        names_not_found: Set[str] = set()
        take_callback_partial = partial(
            self._take_callback, names_not_found_handler=names_not_found.add
        )

        try:
            expression = parse_boolean_expr(spec.func, take_callback_partial, operator_mapping)
        except SyntaxError as err:
            raise InvalidDefinition(
                _("Failed to parse boolean expression '{}'").format(spec.func)
            ) from err
        if not expression or names_not_found:
            spec.names_not_found = names_not_found
            return

        yield expression.unique_key, lambda: expression

    def search(self, spec: "CallbackSpec"):
        if spec.reference is SpecReference.NAME:
            yield from self.search_name(spec.attr_name)
        elif spec.reference is SpecReference.CALLABLE:
            yield from self._search_callable(spec)
        elif spec.reference is SpecReference.PROPERTY:
            yield from self._search_property(spec)
        else:  # never reached here from tests but put an exception for safety. pragma: no cover
            raise ValueError(f"Invalid reference {spec.reference}")

    def _search_property(self, spec):
        # if the attr is a property, we'll try to find the object that has the
        # property on the configs
        attr_name = spec.attr_name
        if attr_name not in self.all_attrs:
            return
        for listener in self.items:
            func = getattr(type(listener.obj), attr_name, None)
            if func is not None and func is spec.func:
                yield (
                    listener.build_key(attr_name),
                    partial(attr_method, attr_name, listener.obj),
                )
                return

    def _search_callable(self, spec):
        # if the attr is an unbounded method, we'll try to find the bounded method
        # on the self
        if not spec.is_bounded:
            for listener in self.items:
                func = getattr(listener.obj, spec.attr_name, None)
                if func is not None and func.__func__ is spec.func:
                    yield listener.build_key(spec.attr_name), partial(callable_method, func)
                    return

        yield f"{spec.attr_name}@None", partial(callable_method, spec.func)

    def search_name(self, name):
        for listener in self.items:
            if name not in listener.all_attrs:
                continue

            key = listener.build_key(name)
            func = getattr(listener.obj, name)
            if not callable(func):
                yield key, partial(attr_method, name, listener.obj)
                continue

            if isinstance(func, Event):
                yield key, partial(event_method, func)
                continue

            yield key, partial(callable_method, func)


def callable_method(a_callable) -> Callable:
    sig = SignatureAdapter.from_callable(a_callable)
    sig_bind_expected = sig.bind_expected

    metadata_to_copy = a_callable.func if isinstance(a_callable, partial) else a_callable

    if sig.is_coroutine:

        async def signature_adapter(*args: Any, **kwargs: Any) -> Any:
            ba = sig_bind_expected(*args, **kwargs)
            return await a_callable(*ba.args, **ba.kwargs)
    else:

        def signature_adapter(*args: Any, **kwargs: Any) -> Any:  # type: ignore[misc]
            ba = sig_bind_expected(*args, **kwargs)
            return a_callable(*ba.args, **ba.kwargs)

    signature_adapter.__name__ = metadata_to_copy.__name__
    signature_adapter.__doc__ = metadata_to_copy.__doc__
    signature_adapter.is_coroutine = sig.is_coroutine  # type: ignore[attr-defined]

    return signature_adapter


def attr_method(attribute, obj) -> Callable:
    getter = attrgetter(attribute)

    def method(*args, **kwargs):
        return getter(obj)

    method.__name__ = attribute
    return method


def event_method(func) -> Callable:
    def method(*args, **kwargs):
        kwargs.pop("machine", None)
        return func(*args, **kwargs)

    return method


def resolver_factory_from_objects(*objects: Tuple[Any, ...]):
    return Listeners.from_listeners(Listener.from_obj(o) for o in objects)



"""
    ulid/time/base
    ~~~~~~~~~~~~~~

    Contains time abstract providers.
"""
import abc

from ... import hints


class Provider(metaclass=abc.ABCMeta):
    """
    Abstract class that defines providers for current time values.
    """
    @abc.abstractmethod
    def milliseconds(self) -> hints.Int:
        """
        Get the current time since unix epoch in milliseconds.

        :return: Epoch timestamp in milliseconds.
        :rtype: :class:`~int`
        """
        raise NotImplementedError('Method must be implemented by derived class')

    @abc.abstractmethod
    def microseconds(self) -> hints.Int:
        """
        Get the current time since unix epoch in microseconds.

        :return: Epoch timestamp in microseconds.
        :rtype: :class:`~int`
        """
        raise NotImplementedError('Method must be implemented by derived class')



"""s3 result store backend."""

from kombu.utils.encoding import bytes_to_str

from celery.exceptions import ImproperlyConfigured

from .base import KeyValueStoreBackend

try:
    import boto3
    import botocore
except ImportError:
    boto3 = None
    botocore = None


__all__ = ('S3Backend',)


class S3Backend(KeyValueStoreBackend):
    """An S3 task result store.

    Raises:
        celery.exceptions.ImproperlyConfigured:
            if module :pypi:`boto3` is not available,
            if the :setting:`aws_access_key_id` or
            setting:`aws_secret_access_key` are not set,
            or it the :setting:`bucket` is not set.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if not boto3 or not botocore:
            raise ImproperlyConfigured('You must install boto3'
                                       'to use s3 backend')
        conf = self.app.conf

        self.endpoint_url = conf.get('s3_endpoint_url', None)
        self.aws_region = conf.get('s3_region', None)

        self.aws_access_key_id = conf.get('s3_access_key_id', None)
        self.aws_secret_access_key = conf.get('s3_secret_access_key', None)

        self.bucket_name = conf.get('s3_bucket', None)
        if not self.bucket_name:
            raise ImproperlyConfigured('Missing bucket name')

        self.base_path = conf.get('s3_base_path', None)

        self._s3_resource = self._connect_to_s3()

    def _get_s3_object(self, key):
        key_bucket_path = self.base_path + key if self.base_path else key
        return self._s3_resource.Object(self.bucket_name, key_bucket_path)

    def get(self, key):
        key = bytes_to_str(key)
        s3_object = self._get_s3_object(key)
        try:
            s3_object.load()
            data = s3_object.get()['Body'].read()
            return data if self.content_encoding == 'binary' else data.decode('utf-8')
        except botocore.exceptions.ClientError as error:
            if error.response['Error']['Code'] == "404":
                return None
            raise error

    def set(self, key, value):
        key = bytes_to_str(key)
        s3_object = self._get_s3_object(key)
        s3_object.put(Body=value)

    def delete(self, key):
        key = bytes_to_str(key)
        s3_object = self._get_s3_object(key)
        s3_object.delete()

    def _connect_to_s3(self):
        session = boto3.Session(
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            region_name=self.aws_region
        )
        if session.get_credentials() is None:
            raise ImproperlyConfigured('Missing aws s3 creds')
        return session.resource('s3', endpoint_url=self.endpoint_url)



# Copyright 2014-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

"""Class to monitor a MongoDB server on a background thread."""

from __future__ import annotations

import asyncio
import atexit
import logging
import time
import weakref
from typing import TYPE_CHECKING, Any, Optional

from pymongo import common, periodic_executor
from pymongo._csot import MovingMinimum
from pymongo.asynchronous.srv_resolver import _SrvResolver
from pymongo.errors import NetworkTimeout, _OperationCancelled
from pymongo.hello import Hello
from pymongo.lock import _async_create_lock
from pymongo.logger import _SDAM_LOGGER, _debug_log, _SDAMStatusMessage
from pymongo.periodic_executor import _shutdown_executors
from pymongo.pool_options import _is_faas
from pymongo.read_preferences import MovingAverage
from pymongo.server_description import ServerDescription

if TYPE_CHECKING:
    from pymongo.asynchronous.pool import (  # type: ignore[attr-defined]
        AsyncConnection,
        Pool,
        _CancellationContext,
    )
    from pymongo.asynchronous.settings import TopologySettings
    from pymongo.asynchronous.topology import Topology

_IS_SYNC = False


def _sanitize(error: Exception) -> None:
    """PYTHON-2433 Clear error traceback info."""
    error.__traceback__ = None
    error.__context__ = None
    error.__cause__ = None


def _monotonic_duration(start: float) -> float:
    """Return the duration since the given start time.

    Accounts for buggy platforms where time.monotonic() is not monotonic.
    See PYTHON-4600.
    """
    return max(0.0, time.monotonic() - start)


class MonitorBase:
    def __init__(self, topology: Topology, name: str, interval: int, min_interval: float):
        """Base class to do periodic work on a background thread.

        The background thread is signaled to stop when the Topology or
        this instance is freed.
        """

        # We strongly reference the executor and it weakly references us via
        # this closure. When the monitor is freed, stop the executor soon.
        async def target() -> bool:
            monitor = self_ref()
            if monitor is None:
                return False  # Stop the executor.
            await monitor._run()  # type:ignore[attr-defined]
            return True

        executor = periodic_executor.AsyncPeriodicExecutor(
            interval=interval, min_interval=min_interval, target=target, name=name
        )

        self._executor = executor

        def _on_topology_gc(dummy: Optional[Topology] = None) -> None:
            # This prevents GC from waiting 10 seconds for hello to complete
            # See test_cleanup_executors_on_client_del.
            monitor = self_ref()
            if monitor:
                monitor.gc_safe_close()

        # Avoid cycles. When self or topology is freed, stop executor soon.
        self_ref = weakref.ref(self, executor.close)
        self._topology = weakref.proxy(topology, _on_topology_gc)
        _register(self)

    def open(self) -> None:
        """Start monitoring, or restart after a fork.

        Multiple calls have no effect.
        """
        self._executor.open()

    def gc_safe_close(self) -> None:
        """GC safe close."""
        self._executor.close()

    async def close(self) -> None:
        """Close and stop monitoring.

        open() restarts the monitor after closing.
        """
        self.gc_safe_close()

    async def join(self) -> None:
        """Wait for the monitor to stop."""
        await self._executor.join()

    def request_check(self) -> None:
        """If the monitor is sleeping, wake it soon."""
        self._executor.wake()


class Monitor(MonitorBase):
    def __init__(
        self,
        server_description: ServerDescription,
        topology: Topology,
        pool: Pool,
        topology_settings: TopologySettings,
    ):
        """Class to monitor a MongoDB server on a background thread.

        Pass an initial ServerDescription, a Topology, a Pool, and
        TopologySettings.

        The Topology is weakly referenced. The Pool must be exclusive to this
        Monitor.
        """
        super().__init__(
            topology,
            "pymongo_server_monitor_task",
            topology_settings.heartbeat_frequency,
            common.MIN_HEARTBEAT_INTERVAL,
        )
        self._server_description = server_description
        self._pool = pool
        self._settings = topology_settings
        self._listeners = self._settings._pool_options._event_listeners
        self._publish = self._listeners is not None and self._listeners.enabled_for_server_heartbeat
        self._cancel_context: Optional[_CancellationContext] = None
        self._conn_id: Optional[int] = None
        self._rtt_monitor = _RttMonitor(
            topology,
            topology_settings,
            topology._create_pool_for_monitor(server_description.address),
        )
        if topology_settings.server_monitoring_mode == "stream":
            self._stream = True
        elif topology_settings.server_monitoring_mode == "poll":
            self._stream = False
        else:
            self._stream = not _is_faas()

    def cancel_check(self) -> None:
        """Cancel any concurrent hello check.

        Note: this is called from a weakref.proxy callback and MUST NOT take
        any locks.
        """
        context = self._cancel_context
        if context:
            # Note: we cannot close the socket because doing so may cause
            # concurrent reads/writes to hang until a timeout occurs
            # (depending on the platform).
            context.cancel()

    async def _start_rtt_monitor(self) -> None:
        """Start an _RttMonitor that periodically runs ping."""
        # If this monitor is closed directly before (or during) this open()
        # call, the _RttMonitor will not be closed. Checking if this monitor
        # was closed directly after resolves the race.
        self._rtt_monitor.open()
        if self._executor._stopped:
            await self._rtt_monitor.close()

    def gc_safe_close(self) -> None:
        self._executor.close()
        self._rtt_monitor.gc_safe_close()
        self.cancel_check()

    async def join(self) -> None:
        await asyncio.gather(
            self._executor.join(), self._rtt_monitor.join(), return_exceptions=True
        )  # type: ignore[func-returns-value]

    async def close(self) -> None:
        self.gc_safe_close()
        await self._rtt_monitor.close()
        # Increment the generation and maybe close the socket. If the executor
        # thread has the socket checked out, it will be closed when checked in.
        await self._reset_connection()

    async def _reset_connection(self) -> None:
        # Clear our pooled connection.
        await self._pool.reset()

    async def _run(self) -> None:
        try:
            prev_sd = self._server_description
            try:
                self._server_description = await self._check_server()
            except _OperationCancelled as exc:
                _sanitize(exc)
                # Already closed the connection, wait for the next check.
                self._server_description = ServerDescription(
                    self._server_description.address, error=exc
                )
                if prev_sd.is_server_type_known:
                    # Immediately retry since we've already waited 500ms to
                    # discover that we've been cancelled.
                    self._executor.skip_sleep()
                return

            # Update the Topology and clear the server pool on error.
            await self._topology.on_change(
                self._server_description,
                reset_pool=self._server_description.error,
                interrupt_connections=isinstance(self._server_description.error, NetworkTimeout),
            )

            if self._stream and (
                self._server_description.is_server_type_known
                and self._server_description.topology_version
            ):
                await self._start_rtt_monitor()
                # Immediately check for the next streaming response.
                self._executor.skip_sleep()

            if self._server_description.error and prev_sd.is_server_type_known:
                # Immediately retry on network errors.
                self._executor.skip_sleep()
        except ReferenceError:
            # Topology was garbage-collected.
            await self.close()
        finally:
            if self._executor._stopped:
                await self._rtt_monitor.close()

    async def _check_server(self) -> ServerDescription:
        """Call hello or read the next streaming response.

        Returns a ServerDescription.
        """
        self._conn_id = None
        start = time.monotonic()
        try:
            return await self._check_once()
        except ReferenceError:
            raise
        except Exception as error:
            _sanitize(error)
            sd = self._server_description
            address = sd.address
            duration = _monotonic_duration(start)
            awaited = bool(self._stream and sd.is_server_type_known and sd.topology_version)
            if self._publish:
                assert self._listeners is not None
                self._listeners.publish_server_heartbeat_failed(address, duration, error, awaited)
            if _SDAM_LOGGER.isEnabledFor(logging.DEBUG):
                _debug_log(
                    _SDAM_LOGGER,
                    message=_SDAMStatusMessage.HEARTBEAT_FAIL,
                    topologyId=self._topology._topology_id,
                    serverHost=address[0],
                    serverPort=address[1],
                    awaited=awaited,
                    durationMS=duration * 1000,
                    failure=error,
                    driverConnectionId=self._conn_id,
                )
            await self._reset_connection()
            if isinstance(error, _OperationCancelled):
                raise
            await self._rtt_monitor.reset()
            # Server type defaults to Unknown.
            return ServerDescription(address, error=error)

    async def _check_once(self) -> ServerDescription:
        """A single attempt to call hello.

        Returns a ServerDescription, or raises an exception.
        """
        address = self._server_description.address
        sd = self._server_description

        # XXX: "awaited" could be incorrectly set to True in the rare case
        # the pool checkout closes and recreates a connection.
        awaited = bool(
            self._pool.conns and self._stream and sd.is_server_type_known and sd.topology_version
        )
        if self._publish:
            assert self._listeners is not None
            self._listeners.publish_server_heartbeat_started(address, awaited)

        if self._cancel_context and self._cancel_context.cancelled:
            await self._reset_connection()
        async with self._pool.checkout() as conn:
            if _SDAM_LOGGER.isEnabledFor(logging.DEBUG):
                _debug_log(
                    _SDAM_LOGGER,
                    message=_SDAMStatusMessage.HEARTBEAT_START,
                    topologyId=self._topology._topology_id,
                    driverConnectionId=conn.id,
                    serverConnectionId=conn.server_connection_id,
                    serverHost=address[0],
                    serverPort=address[1],
                    awaited=awaited,
                )

            self._cancel_context = conn.cancel_context
            # Record the connection id so we can later attach it to the failed log message.
            self._conn_id = conn.id
            response, round_trip_time = await self._check_with_socket(conn)
            if not response.awaitable:
                await self._rtt_monitor.add_sample(round_trip_time)

            avg_rtt, min_rtt = await self._rtt_monitor.get()
            sd = ServerDescription(address, response, avg_rtt, min_round_trip_time=min_rtt)
            if self._publish:
                assert self._listeners is not None
                self._listeners.publish_server_heartbeat_succeeded(
                    address, round_trip_time, response, response.awaitable
                )
            if _SDAM_LOGGER.isEnabledFor(logging.DEBUG):
                _debug_log(
                    _SDAM_LOGGER,
                    message=_SDAMStatusMessage.HEARTBEAT_SUCCESS,
                    topologyId=self._topology._topology_id,
                    driverConnectionId=conn.id,
                    serverConnectionId=conn.server_connection_id,
                    serverHost=address[0],
                    serverPort=address[1],
                    awaited=awaited,
                    durationMS=round_trip_time * 1000,
                    reply=response.document,
                )
            return sd

    async def _check_with_socket(self, conn: AsyncConnection) -> tuple[Hello, float]:
        """Return (Hello, round_trip_time).

        Can raise ConnectionFailure or OperationFailure.
        """
        start = time.monotonic()
        if conn.more_to_come:
            # Read the next streaming hello (MongoDB 4.4+).
            response = Hello(await conn._next_reply(), awaitable=True)
        elif (
            self._stream and conn.performed_handshake and self._server_description.topology_version
        ):
            # Initiate streaming hello (MongoDB 4.4+).
            response = await conn._hello(
                self._server_description.topology_version,
                self._settings.heartbeat_frequency,
            )
        else:
            # New connection handshake or polling hello (MongoDB <4.4).
            response = await conn._hello(None, None)
        duration = _monotonic_duration(start)
        return response, duration


class SrvMonitor(MonitorBase):
    def __init__(self, topology: Topology, topology_settings: TopologySettings):
        """Class to poll SRV records on a background thread.

        Pass a Topology and a TopologySettings.

        The Topology is weakly referenced.
        """
        super().__init__(
            topology,
            "pymongo_srv_polling_thread",
            common.MIN_SRV_RESCAN_INTERVAL,
            topology_settings.heartbeat_frequency,
        )
        self._settings = topology_settings
        self._seedlist = self._settings._seeds
        assert isinstance(self._settings.fqdn, str)
        self._fqdn: str = self._settings.fqdn
        self._startup_time = time.monotonic()

    async def _run(self) -> None:
        # Don't poll right after creation, wait 60 seconds first
        if time.monotonic() < self._startup_time + common.MIN_SRV_RESCAN_INTERVAL:
            return
        seedlist = await self._get_seedlist()
        if seedlist:
            self._seedlist = seedlist
            try:
                await self._topology.on_srv_update(self._seedlist)
            except ReferenceError:
                # Topology was garbage-collected.
                await self.close()

    async def _get_seedlist(self) -> Optional[list[tuple[str, Any]]]:
        """Poll SRV records for a seedlist.

        Returns a list of ServerDescriptions.
        """
        try:
            resolver = _SrvResolver(
                self._fqdn,
                self._settings.pool_options.connect_timeout,
                self._settings.srv_service_name,
            )
            seedlist, ttl = await resolver.get_hosts_and_min_ttl()
            if len(seedlist) == 0:
                # As per the spec: this should be treated as a failure.
                raise Exception
        except Exception:
            # As per the spec, upon encountering an error:
            # - An error must not be raised
            # - SRV records must be rescanned every heartbeatFrequencyMS
            # - Topology must be left unchanged
            self.request_check()
            return None
        else:
            self._executor.update_interval(max(ttl, common.MIN_SRV_RESCAN_INTERVAL))
            return seedlist


class _RttMonitor(MonitorBase):
    def __init__(self, topology: Topology, topology_settings: TopologySettings, pool: Pool):
        """Maintain round trip times for a server.

        The Topology is weakly referenced.
        """
        super().__init__(
            topology,
            "pymongo_server_rtt_task",
            topology_settings.heartbeat_frequency,
            common.MIN_HEARTBEAT_INTERVAL,
        )

        self._pool = pool
        self._moving_average = MovingAverage()
        self._moving_min = MovingMinimum()
        self._lock = _async_create_lock()

    async def close(self) -> None:
        self.gc_safe_close()
        # Increment the generation and maybe close the socket. If the executor
        # thread has the socket checked out, it will be closed when checked in.
        await self._pool.reset()

    async def add_sample(self, sample: float) -> None:
        """Add a RTT sample."""
        async with self._lock:
            self._moving_average.add_sample(sample)
            self._moving_min.add_sample(sample)

    async def get(self) -> tuple[Optional[float], float]:
        """Get the calculated average, or None if no samples yet and the min."""
        async with self._lock:
            return self._moving_average.get(), self._moving_min.get()

    async def reset(self) -> None:
        """Reset the average RTT."""
        async with self._lock:
            self._moving_average.reset()
            self._moving_min.reset()

    async def _run(self) -> None:
        try:
            # NOTE: This thread is only run when using the streaming
            # heartbeat protocol (MongoDB 4.4+).
            # XXX: Skip check if the server is unknown?
            rtt = await self._ping()
            await self.add_sample(rtt)
        except ReferenceError:
            # Topology was garbage-collected.
            await self.close()
        except Exception:
            await self._pool.reset()

    async def _ping(self) -> float:
        """Run a "hello" command and return the RTT."""
        async with self._pool.checkout() as conn:
            if self._executor._stopped:
                raise Exception("_RttMonitor closed")
            start = time.monotonic()
            await conn.hello()
            return _monotonic_duration(start)


# Close monitors to cancel any in progress streaming checks before joining
# executor threads. For an explanation of how this works see the comment
# about _EXECUTORS in periodic_executor.py.
_MONITORS = set()


def _register(monitor: MonitorBase) -> None:
    ref = weakref.ref(monitor, _unregister)
    _MONITORS.add(ref)


def _unregister(monitor_ref: weakref.ReferenceType[MonitorBase]) -> None:
    _MONITORS.remove(monitor_ref)


def _shutdown_monitors() -> None:
    if _MONITORS is None:
        return

    # Copy the set. Closing monitors removes them.
    monitors = list(_MONITORS)

    # Close all monitors.
    for ref in monitors:
        monitor = ref()
        if monitor:
            monitor.gc_safe_close()

    monitor = None


def _shutdown_resources() -> None:
    # _shutdown_monitors/_shutdown_executors may already be GC'd at shutdown.
    shutdown = _shutdown_monitors
    if shutdown:  # type:ignore[truthy-function]
        shutdown()
    shutdown = _shutdown_executors
    if shutdown:  # type:ignore[truthy-function]
        shutdown()


if _IS_SYNC:
    atexit.register(_shutdown_resources)



from typing import Callable

from .callbacks import CallbackGroup


class AddCallbacksMixin:
    def _add_callback(self, callback, grouper: CallbackGroup, is_event=False, **kwargs):
        raise NotImplementedError

    def __call__(self, f):
        return self._add_callback(f, CallbackGroup.ON, is_event=True)

    def before(self, f: Callable):
        """Adds a ``before`` :ref:`transition actions` callback to every :ref:`transition` in the
        :ref:`TransitionList` instance.

        Args:
            f: The ``before`` :ref:`transition actions` callback function to be added.

        Returns:
            The `f` callable.
        """
        return self._add_callback(f, CallbackGroup.BEFORE)

    def after(self, f: Callable):
        """Adds a ``after`` :ref:`transition actions` callback to every :ref:`transition` in the
        :ref:`TransitionList` instance.

        Args:
            f: The ``after`` :ref:`transition actions` callback function to be added.

        Returns:
            The `f` callable.
        """
        return self._add_callback(f, CallbackGroup.AFTER)

    def on(self, f: Callable):
        """Adds a ``on`` :ref:`transition actions` callback to every :ref:`transition` in the
        :ref:`TransitionList` instance.

        Args:
            f: The ``on`` :ref:`transition actions` callback function to be added.

        Returns:
            The `f` callable.
        """
        return self._add_callback(f, CallbackGroup.ON)

    def cond(self, f: Callable):
        """Adds a ``cond`` :ref:`guards` callback to every :ref:`transition` in the
        :ref:`TransitionList` instance.

        Args:
            f: The ``cond`` :ref:`guards` callback function to be added.

        Returns:
            The `f` callable.
        """
        return self._add_callback(f, CallbackGroup.COND, expected_value=True)

    def unless(self, f: Callable):
        """Adds a ``unless`` :ref:`guards` callback with expected value ``False`` to every
        :ref:`transition` in the :ref:`TransitionList` instance.

        Args:
            f: The ``unless`` :ref:`guards` callback function to be added.

        Returns:
            The `f` callable.
        """
        return self._add_callback(f, CallbackGroup.COND, expected_value=False)

    def validators(self, f: Callable):
        """Adds a :ref:`validators` callback to the :ref:`TransitionList` instance.

        Args:
            f: The ``validators`` callback function to be added.
        Returns:
            The callback function.

        """
        return self._add_callback(f, CallbackGroup.VALIDATOR)



import datetime
import enum
import json
from functools import cached_property, cache
from typing import Optional, Any

from loguru import logger
from pydantic import Field, BaseModel, ConfigDict
from statemachine import StateMachine, State
from statemachine.event_data import EventData

from genie_flow.model.dialogue import DialogueElement, DialogueFormat
from genie_flow.model.secondary_store import SecondaryStore
from genie_flow.model.template import CompositeTemplateType
from genie_flow.model.versioned import VersionedModel


class StateType(enum.IntEnum):
    USER = 0
    INVOKER = 1

    @property
    def as_actor(self) -> str:
        match self:
            case StateType.INVOKER:
                return "assistant"
            case StateType.USER:
                return "user"
            case _:
                raise ValueError("Unknown State Type")


class DialoguePersistence(enum.IntEnum):
    NONE = 0
    RAW = 1
    RENDERED = 2


class GenieModel(VersionedModel):
    """
    The base model for all models that will carry data in the dialogue. Contains the attributes
    that are required and expected by the `GenieStateMachine` such as `state` and `session_id`/

    This class also carries the dialogue - a list of `DialogueElement`s of the chat so far.

    And it carries a number of state-dependent attributes that are important to the progress of
    the dialogue, such as `running_task_id` which indicates if there is a currently running task,
    as well as `actor` and `actor_text`, both indicators for the most recent interaction.

    This class is a subclass of the pydantic_redis `Model` class, which makes it possible to
    persist the values into Reids and retrieve it again by its primary key. The attribute
    `_primary_key_field` is used to determine the name of the primary key.
    """

    state: str | int | None = Field(
        None,
        description="The current state that this model is in, represented by the state's value",
    )
    session_id: str = Field(
        description="The ID of the session this data model object belongs to."
    )
    source_type: Optional[StateType] = Field(
        default=None,
        description="The type of state that the most recent transition is from",
    )
    target_type: Optional[StateType] = Field(
        default=None,
        description="The type of state that the most recent transition is to",
    )
    dialogue_persistence: Optional[DialoguePersistence] = Field(
        default=None,
        description="Indicator to how to add most recent actor input to the dialogue",
    )
    dialogue: list[DialogueElement] = Field(
        default_factory=list,
        description="The list of dialogue elements that have been used in the dialogue so far",
    )
    task_error: Optional[str] = Field(
        default=None,
        description="The error message returned from a running task",
    )
    actor: Optional[str] = Field(
        None,
        description="The actor that has created the current input",
    )
    actor_input: str = Field(
        "",
        description="the most recent received input from the actor",
    )
    secondary_storage: SecondaryStore = Field(
        default_factory=SecondaryStore,
        description="A dictionary that can be used to store secondary information about the session",
    )

    @property
    def render_data(self) -> dict[str, Any]:
        """
        Returns a dictionary containing all data that can be used to render a template.

        **NOTE** We are using the `serialize_as_any` flag here to make sure that properties
        in the `secondary_storage` are also included.

        It will contain:
        - "state_id": The ID of the current state of the state machine
        - "current_datetime": The ISO 8601 formatted current date and time in UTC
        - "dialogue" The string output of the current dialogue
        - all keys and values of the machine's current model
        """
        render_data = self.model_dump(serialize_as_any=True)
        try:
            parsed_json = json.loads(self.actor_input)
        except json.JSONDecodeError:
            parsed_json = None

        render_data.update(
            {
                "parsed_actor_input": parsed_json,
                "state_id": self.state,
                "current_datetime": datetime.datetime.now(datetime.timezone.utc).isoformat(),
                "chat_history": str(self.format_dialogue(DialogueFormat.YAML)),
            }
        )
        return render_data

    @property
    def has_errors(self) -> bool:
        return self.task_error is not None

    @classmethod
    def get_state_machine_class(cls) -> type["GenieStateMachine"]:
        """
        Property that returns the class of the state machine that this model should be
        managed by.
        """
        raise NotImplementedError()

    @property
    def current_response(self) -> Optional[DialogueElement]:
        """
        Return the most recent `DialogueElement` from the dialogue list.
        """
        return self.dialogue[-1] if len(self.dialogue) > 0 else None

    def format_dialogue(self, target_format: DialogueFormat) -> str:
        """
        Apply the given target format to the dialogue of this instance.
        """
        return DialogueFormat.format(self.dialogue, target_format)

    def add_dialogue_element(self, actor: str, actor_text: str):
        """
        Add a given actor and actor text to the dialogue.
        :param actor: the name of the actor
        :param actor_text: the actor text
        """
        element = DialogueElement(actor=actor, actor_text=actor_text)
        self.dialogue.append(element)

    def record_dialogue_element(self):
        """
        Record the current `actor` and `actor_input` into the dialogue.
       """
        self.add_dialogue_element(self.actor, self.actor_input)


class GenieStateMachine(StateMachine):
    """
    A State Machine class that is able to manage an AI driven dialogue and extract information
    from it. The extracted information is stored in an accompanying data model (based on the
    `GenieModel` class).
    """

    # EVENTS that need to be specified
    user_input: Any = None
    ai_extraction: Any = None
    advance: Any = None

    # TEMPLATE mapping that needs to be specified
    templates: dict[str, CompositeTemplateType] = dict()

    def __init__(
        self,
        model: GenieModel,
    ):
        self.current_template: Optional[CompositeTemplateType] = None
        super(GenieStateMachine, self).__init__(model=model)

    @property
    def render_data(self) -> dict[str, Any]:
        """
        Returns a dictionary containing all data that can be used to render a template.

        It will contain:
        - "state_id": The ID of the current state of the state machine
        - "state_name": The name of the current state of the state machine
        - "dialogue" The string output of the current dialogue
        - all keys and values of the machine's current model
        """
        render_data = self.model.model_dump()
        try:
            parsed_json = json.loads(self.model.actor_input)
        except json.JSONDecodeError:
            parsed_json = None

        render_data.update(
            {
                "parsed_actor_input": parsed_json,
                "state_id": self.current_state.id,
                "state_name": self.current_state.name,
                "chat_history": str(self.model.format_dialogue(DialogueFormat.YAML)),
            }
        )
        return render_data

    def get_template_for_state(self, state: State) -> CompositeTemplateType:
        """
        Retrieve the template for a given state. Raises an exception if the given
        state does not have a template defined.

        :param state: The state for which to retrieve the template for
        :return: The template for the given state
        :raises AttributeError: If this ob`ject does not have an attribute that carries the templates
        :raises KeyError: If there is no template defined for the given state
        """
        try:
            return self.templates[state.id]
        except KeyError:
            logger.error(f"No template for state {state.id}")
            raise

    # VALIDATIONS AND CONDITIONS
    def is_valid_response(self, event_data: EventData):
        logger.debug(f"is valid response {event_data.args}")
        return all(
            [
                event_data.args is not None,
                len(event_data.args) > 0,
                event_data.args[0] is not None,
                event_data.args[0] != "",
            ]
        )



"""See ``with_plugins()``."""


from importlib.metadata import entry_points as im_entry_points
import os
import sys
import traceback

import click


__version__ = '2.0dev'
__author__ = 'Kevin Wurster, Sean Gillies'
__email__ = 'wursterk@gmail.com, sean.gillies@gmail.com'
__source__ = 'https://github.com/click-contrib/click-plugins'
__license__ = '''
New BSD License

Copyright (c) 2015-2023, Kevin D. Wurster, Sean C. Gillies
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither click-plugins nor the names of its contributors may not be used to
  endorse or promote products derived from this software without specific prior
  written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''


def with_plugins(entry_points):

    """Decorator for loading plugins.

    Each entry point must point to a ``click.Command()`` object. An entry
    point producing an exception during loading will be wrapped in a
    ``BrokenCommand()``.

    Entry points can be loaded from multiple sources by decorating a command
    with multiple ``with_plugins()`` using different configurations.

    >>> from importlib.metadata import entry_points
    >>> import click
    >>> from click_plugins import with_plugins
    >>>
    >>> @with_plugins(entry_points(group='entry_point.name'))
    >>> @click.group()
    >>> def cli():
    ...     '''Commandline interface for something.'''
    >>>
    >>> @cli.command()
    >>> @click.argument('arg')
    >>> def subcommand(arg):
    ...     '''A subcommand for something else'''

    :param iterable or str entry_points:
        A string identifying a group of entry points to load or an iterable
        object producing ``importlib.metadata.EntryPoint()`` objects.

    :rtype click.Group:
    """

    def decorator(group):
        if not isinstance(group, click.Group):
            raise TypeError(
                f"plugins can only be attached to an instance of"
                f" 'click.Group()' not: {repr(group)}")

        # Load 'EntryPoint()' objects.
        if isinstance(entry_points, str):
            all_entry_points = im_entry_points(group=entry_points)

        # Receiving an iterable of 'EntryPoints()'.
        else:
            all_entry_points = entry_points

        for ep in all_entry_points:

            try:
                group.add_command(ep.load())

            # Catch all exceptions (technically not 'BaseException') and
            # instead register a special 'BrokenCommand()'. Otherwise, a single
            # plugin that fails to load and/or register will make the CLI
            # inoperable. 'BrokenCommand()' explains the situation to users.
            except Exception as e:
                group.add_command(BrokenCommand(ep, e))

        return group

    return decorator


class BrokenCommand(click.Command):

    """Represents a plugin ``click.Command()`` that failed to load.

    Can be executed just like a ``click.Command()``, but prints information
    for debugging and exits with an error code.
    """

    def __init__(self, entry_point, exception):

        """
        :param importlib.metadata.EntryPoint entry_point:
            Entry point that failed to load.
        :param Exception exception:
            Raised when attempting to load the entry point associated with
            this instance.
        """

        super().__init__(entry_point.name)

        # There are several ways to get a traceback from an exception, but
        # 'TracebackException()' seems to be the most portable across actively
        # supported versions of Python.
        tbe = traceback.TracebackException.from_exception(exception)

        # A message for '$ cli command --help'. Contains full traceback and a
        # helpful note. The intention is to nudge users to figure out which
        # project should get a bug report since users are likely to report the
        # issue to the developers of the CLI utility they are directly
        # interacting with. These are not necessarily the right developers.
        self.help = (
            "{ls}ERROR: entry point '{module}:{name}' could not be loaded."
            " Contact its author for help.{ls}{ls}{tb}").format(
            module=_module(entry_point),
            name=entry_point.name,
            ls=os.linesep,
            tb=''.join(tbe.format())
        )

        # Replace the broken command's summary with a warning about how it
        # was not loaded successfully. The idea is that '$ cli --help' should
        # include a clear indicator that a subcommand is not functional, and
        # a little hint for what to do about it. U+2020 is a "dagger", whose
        # modern use typically indicates a footnote.
        self.short_help = (
            f"\u2020 Warning: could not load plugin. Invoke command with"
            f" '--help' for traceback."
        )

    def invoke(self, ctx):

        """Print traceback and debugging message.

        :param click.Context ctx:
            Active context.
        """

        click.echo(self.help, color=ctx.color, err=True)
        ctx.exit(1)

    def parse_args(self, ctx, args):

        """Pass arguments along without parsing.

        :param click.Context ctx:
            Active context.
        :param list args:
            List of command line arguments.
        """

        # Do not attempt to parse these arguments. We do not know why the
        # entry point failed to load, but it is reasonable to assume that
        # argument parsing will not work. Ultimately the goal is to get the
        # 'Command.invoke()' method (overloaded in this class) to execute
        # and provide the user with a bit of debugging information.

        return args


def _module(ep):

    """Module name for a given entry point.

    Parameters
    ----------
    ep : importlib.metadata.EntryPoint
        Determine parent module for this entry point.

    Returns
    -------
    str
    """

    if sys.version_info >= (3, 10):
        module = ep.module

    else:
        # From 'importlib.metadata.EntryPoint.module'.
        match = ep.pattern.match(ep.value)
        module = match.group('module')

    return module


# def _entry_points(**params):
#
#     eps = entry_points()
#
#     # 'entry_points()' returned an 'importlib.metadata.EntryPoints()'.
#     if sys.version_info >= (3, 9):
#
#         # If 'params' is empty it will produce a 'DeprecationWarning' on some
#         # versions of Python. This is due to a bug in 'select()' that returns
#         # 'SelectableGroup()' instead of 'EntryPoints()' in this
#         eps = eps.select(**params)
#
#     # Got a 'dict()'. Keys are groups, and values are tuples.
#     else:
#         mapping = eps
#         if list(params.keys()) == ['group']:
#             grp = params['group']
#             eps = mapping.get(grp, tuple())
#         else:
#             raise ValueError(f"must filter on 'group': {params}")
#
#     return eps



from collections import deque
from threading import Lock
from typing import TYPE_CHECKING
from weakref import proxy

from ..event import BoundEvent
from ..event_data import TriggerData
from ..state import State
from ..transition import Transition

if TYPE_CHECKING:
    from ..statemachine import StateMachine


class BaseEngine:
    def __init__(self, sm: "StateMachine", rtc: bool = True):
        self.sm: StateMachine = proxy(sm)
        self._external_queue: deque = deque()
        self._sentinel = object()
        self._rtc = rtc
        self._processing = Lock()

    def put(self, trigger_data: TriggerData):
        """Put the trigger on the queue without blocking the caller."""
        self._external_queue.append(trigger_data)

    def start(self):
        if self.sm.current_state_value is not None:
            return

        trigger_data = TriggerData(
            machine=self.sm,
            event=BoundEvent("__initial__", _sm=self.sm),
        )
        self.put(trigger_data)

    def _initial_transition(self, trigger_data):
        transition = Transition(State(), self.sm._get_initial_state(), event="__initial__")
        transition._specs.clear()
        return transition



from functools import cache

import snappy
from loguru import logger
from pydantic import BaseModel, ConfigDict
from pydantic.main import IncEx
from pydantic_core import from_json


class VersionedModel(BaseModel):
    """
    A base class for models that have a schema version.
    
    This model provides built-in serialization and deserialization capabilities:

    - Serialization: Converts model instances to bytes format with schema versioning and optional compression
    - Deserialization: Restores model instances from bytes, handling schema version compatibility through upgrades
    - Schema versioning: Ensures data compatibility across different model versions through version tracking
    """

    model_config = ConfigDict(json_schema_extra={"schema_version": 0})

    @classmethod
    @property
    @cache
    def schema_version(cls) -> int:
        return int(cls.model_json_schema()["schema_version"])

    def serialize(
        self,
        compression: bool = False,
        include: IncEx | None = None,
        exclude: IncEx | None = None,
    ) -> bytes:
        """
        Creates a serialization of the object. Serialization results in a
        bytes object, containing the schema version number, a compression indicator and
        the serialized version of the model object. All separated by a ':' character.

        Computed fields will be excluded from serialization, except for those explicitly
        included.

        :param compression: a boolean indicating whether to use compression or not
        :param include: fields to include in the serialization
        :param exclude: fields to exclude from the serialization

        :return: a bytes with the serialized version of the model object
        """
        computed_fields_to_exclude = {
            field
            for field in self.__class__.model_computed_fields.keys()
            if include is None or field not in include
        }
        exclude = exclude.union(computed_fields_to_exclude) if exclude else computed_fields_to_exclude
        model_dump = self.model_dump_json(include=include, exclude=exclude)
        if compression:
            payload = snappy.compress(model_dump, encoding="utf-8")
        else:
            payload = model_dump.encode("utf-8")
        compression_flag = b"1" if compression else b"0"

        return b":".join(
            [str(self.schema_version).encode("utf-8"), compression_flag, payload]
        )

    @classmethod
    def upgrade_schema(cls, from_version: int, model_data: dict) -> dict:
        """
        Upgrade the data in the model to the new schema version.

        **NOTE**: This method should be overridden by subclasses.

        :param from_version: the schema version to upgrade from
        :param model_data: the data that should be upgraded
        :return: the data that fits the current schema version
        :raises ValueError: if the schema version cannot be made compatible with the
        current schema version
        """
        logger.error(
            "Cannot deserialize a model with schema version {persisted_version} "
            "into a model with schema version {current_version} "
            "for model class {model_class}",
            persisted_version=int(from_version),
            current_version=cls.schema_version,
            model_class=cls.__name__,
        )
        raise ValueError(f"Schema mis-match when deserializing a {cls.__name__} model")

    @classmethod
    def deserialize(cls, payload: bytes) -> "VersionedModel":
        persisted_version, compression, payload = payload.split(b":", maxsplit=2)
        persisted_version = int(persisted_version.decode("utf-8"))
        model_json = (
            snappy.decompress(payload, decoding="utf-8")
            if compression == b"1"
            else payload.decode("utf-8")
        )

        if persisted_version != cls.schema_version:
            model_data = from_json(model_json)
            model_data = cls.upgrade_schema(persisted_version, model_data)
            return cls.model_validate(model_data)

        return cls.model_validate_json(model_json, by_alias=True, by_name=True)



from abc import ABC, abstractmethod


class GenieInvoker(ABC):
    """
    The super class of all Genie Invokers. The standard interface to invoke large language models,
    database retrievals, etc.

    This is ab abstraction around calls that take a text content and pass that to a lower level
    service for processing. The returned value is always a result string.

    This class is subclassed with specific classes for external services.
    """

    @classmethod
    @abstractmethod
    def from_config(cls, config: dict):
        raise NotImplementedError()

    @abstractmethod
    def invoke(self, content: str) -> str:
        """
        Invoke the underlying service with the supplied content and dialogue.

        :param content: The text content to invoke the underlying service. The format of
        this string is Invoker dependent. Some may simply expect a string, others may
        need to get a structured document as string - for instance a JSON string - that
        incorporates the values that one needs to pass.
        :return: The result string.
        """
        raise NotImplementedError()



"""
    ulid/consts
    ~~~~~~~~~~~

    Contains public API constant values.
"""
from . import ulid

__all__ = ['MIN_TIMESTAMP', 'MAX_TIMESTAMP', 'MIN_RANDOMNESS', 'MAX_RANDOMNESS', 'MIN_ULID', 'MAX_ULID']

#: Minimum possible timestamp value (0).
MIN_TIMESTAMP = ulid.Timestamp(b'\x00\x00\x00\x00\x00\x00')


#: Maximum possible timestamp value (281474976710.655 epoch).
MAX_TIMESTAMP = ulid.Timestamp(b'\xff\xff\xff\xff\xff\xff')


#: Minimum possible randomness value (0).
MIN_RANDOMNESS = ulid.Randomness(b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')


#: Maximum possible randomness value (1208925819614629174706175).
MAX_RANDOMNESS = ulid.Randomness(b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff')


#: Minimum possible ULID value (0).
MIN_ULID = ulid.ULID(MIN_TIMESTAMP.bytes + MIN_RANDOMNESS.bytes)


#: Maximum possible ULID value (340282366920938463463374607431768211455).
MAX_ULID = ulid.ULID(MAX_TIMESTAMP.bytes + MAX_RANDOMNESS.bytes)



from os import PathLike

from celery import Celery
from fastapi import FastAPI

from genie_flow.containers.genieflow import GenieFlowContainer
from genie_flow.environment import GenieEnvironment


class GenieFlow:

    def __init__(self, container: GenieFlowContainer):
        self.container = container

    @classmethod
    def from_yaml(cls, config_file_path: str | PathLike) -> "GenieFlow":
        container = GenieFlowContainer()
        container.config.from_yaml(config_file_path, required=True)
        container.wire(packages=["genie_flow"])
        container.storage.container.wire(packages=["genie_flow.celery"])
        container.permanent_storage.container.wire(packages=["genie_flow.mongo"])
        container.init_resources()

        return cls(container)

    @property
    def genie_environment(self) -> GenieEnvironment:
        return self.container.genie_environment()

    @property
    def fastapi_app(self) -> FastAPI:
        return self.container.fastapi_app()

    @property
    def celery_app(self) -> Celery:
        return self.container.celery_app()



from celery import Celery
from dependency_injector import containers, providers

from genie_flow.app import create_fastapi_app
from genie_flow.containers.core import GenieFlowCoreContainer
from genie_flow.containers.perm_persistence import GenieFlowPermanentPersistenceContainer
from genie_flow.containers.persistence import GenieFlowPersistenceContainer
from genie_flow.celery import CeleryManager
from genie_flow.environment import GenieEnvironment
from genie_flow.model.types import ModelKeyRegistryType
from genie_flow.session import SessionManager
from genie_flow_invoker.factory import InvokerFactory


class GenieFlowContainer(containers.DeclarativeContainer):

    config = providers.Configuration()

    core = providers.Container(
        GenieFlowCoreContainer,
        config=config,
    )

    model_key_registry = providers.Singleton(ModelKeyRegistryType)

    invoker_factory = providers.Factory(
        InvokerFactory,
        config=config.invokers,
    )

    storage = providers.Container(
        GenieFlowPersistenceContainer,
        config=config.persistence,
    )

    permanent_storage = providers.Container(
        GenieFlowPermanentPersistenceContainer,
        config=config.persistence,
    )

    genie_environment = providers.Singleton(
        GenieEnvironment,
        config.genie_environment.template_root_path,
        config.genie_environment.pool_size,
        model_key_registry,
        invoker_factory,
    )

    celery_app = providers.Singleton(
        Celery,
        main="genie_flow",
        broker=config.celery.broker,
        backend=config.celery.backend,
        redis_socket_timeout=config.celery.redis_socket_timeout,
        redis_socket_connect_timeout=config.celery.redis_socket_connect_timeout,
    )

    celery_manager = providers.Singleton(
        CeleryManager,
        celery_app,
        storage.session_lock_manager,
        genie_environment,
        update_mongo_period = config.celery.update_mongo_period or 60.0
    )

    session_manager = providers.Singleton(
        SessionManager,
        session_lock_manager=storage.session_lock_manager,
        model_key_registry=model_key_registry,
        genie_environment=genie_environment,
        celery_manager=celery_manager,
    )

    fastapi_app = providers.Resource(
        create_fastapi_app,
        session_manager=session_manager,
        config=config.api,
        cors_settings=config.cors
    )



"""CouchDB result store backend."""
from kombu.utils.encoding import bytes_to_str
from kombu.utils.url import _parse_url

from celery.exceptions import ImproperlyConfigured

from .base import KeyValueStoreBackend

try:
    import pycouchdb
except ImportError:
    pycouchdb = None

__all__ = ('CouchBackend',)

ERR_LIB_MISSING = """\
You need to install the pycouchdb library to use the CouchDB result backend\
"""


class CouchBackend(KeyValueStoreBackend):
    """CouchDB backend.

    Raises:
        celery.exceptions.ImproperlyConfigured:
            if module :pypi:`pycouchdb` is not available.
    """

    container = 'default'
    scheme = 'http'
    host = 'localhost'
    port = 5984
    username = None
    password = None

    def __init__(self, url=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = url

        if pycouchdb is None:
            raise ImproperlyConfigured(ERR_LIB_MISSING)

        uscheme = uhost = uport = uname = upass = ucontainer = None
        if url:
            _, uhost, uport, uname, upass, ucontainer, _ = _parse_url(url)
            ucontainer = ucontainer.strip('/') if ucontainer else None

        self.scheme = uscheme or self.scheme
        self.host = uhost or self.host
        self.port = int(uport or self.port)
        self.container = ucontainer or self.container
        self.username = uname or self.username
        self.password = upass or self.password

        self._connection = None

    def _get_connection(self):
        """Connect to the CouchDB server."""
        if self.username and self.password:
            conn_string = f'{self.scheme}://{self.username}:{self.password}@{self.host}:{self.port}'
            server = pycouchdb.Server(conn_string, authmethod='basic')
        else:
            conn_string = f'{self.scheme}://{self.host}:{self.port}'
            server = pycouchdb.Server(conn_string)

        try:
            return server.database(self.container)
        except pycouchdb.exceptions.NotFound:
            return server.create(self.container)

    @property
    def connection(self):
        if self._connection is None:
            self._connection = self._get_connection()
        return self._connection

    def get(self, key):
        key = bytes_to_str(key)
        try:
            return self.connection.get(key)['value']
        except pycouchdb.exceptions.NotFound:
            return None

    def set(self, key, value):
        key = bytes_to_str(key)
        data = {'_id': key, 'value': value}
        try:
            self.connection.save(data)
        except pycouchdb.exceptions.Conflict:
            # document already exists, update it
            data = self.connection.get(key)
            data['value'] = value
            self.connection.save(data)

    def mget(self, keys):
        return [self.get(key) for key in keys]

    def delete(self, key):
        self.connection.delete(key)



from typing import TYPE_CHECKING

from ..event_data import EventData
from ..event_data import TriggerData
from ..exceptions import InvalidDefinition
from ..exceptions import TransitionNotAllowed
from ..i18n import _
from .base import BaseEngine

if TYPE_CHECKING:
    from ..statemachine import StateMachine
    from ..transition import Transition


class AsyncEngine(BaseEngine):
    def __init__(self, sm: "StateMachine", rtc: bool = True):
        if not rtc:
            raise InvalidDefinition(_("Only RTC is supported on async engine"))
        super().__init__(sm=sm, rtc=rtc)

    async def activate_initial_state(self):
        """
        Activate the initial state.

        Called automatically on state machine creation from sync code, but in
        async code, the user must call this method explicitly.

        Given how async works on python, there's no built-in way to activate the initial state that
        may depend on async code from the StateMachine.__init__ method.
        """
        return await self.processing_loop()

    async def processing_loop(self):
        """Process event triggers.

        The simplest implementation is the non-RTC (synchronous),
        where the trigger will be run immediately and the result collected as the return.

        .. note::

            While processing the trigger, if others events are generated, they
            will also be processed immediately, so a "nested" behavior happens.

        If the machine is on ``rtc`` model (queued), the event is put on a queue, and only the
        first event will have the result collected.

        .. note::
            While processing the queue items, if others events are generated, they
            will be processed sequentially (and not nested).

        """
        # We make sure that only the first event enters the processing critical section,
        # next events will only be put on the queue and processed by the same loop.
        if not self._processing.acquire(blocking=False):
            return None

        # We will collect the first result as the processing result to keep backwards compatibility
        # so we need to use a sentinel object instead of `None` because the first result may
        # be also `None`, and on this case the `first_result` may be overridden by another result.
        first_result = self._sentinel
        try:
            # Execute the triggers in the queue in FIFO order until the queue is empty
            while self._external_queue:
                trigger_data = self._external_queue.popleft()
                try:
                    result = await self._trigger(trigger_data)
                    if first_result is self._sentinel:
                        first_result = result
                except Exception:
                    # Whe clear the queue as we don't have an expected behavior
                    # and cannot keep processing
                    self._external_queue.clear()
                    raise
        finally:
            self._processing.release()
        return first_result if first_result is not self._sentinel else None

    async def _trigger(self, trigger_data: TriggerData):
        executed = False
        if trigger_data.event == "__initial__":
            transition = self._initial_transition(trigger_data)
            await self._activate(trigger_data, transition)
            return self._sentinel

        state = self.sm.current_state
        for transition in state.transitions:
            if not transition.match(trigger_data.event):
                continue

            executed, result = await self._activate(trigger_data, transition)
            if not executed:
                continue
            break
        else:
            if not self.sm.allow_event_without_transition:
                raise TransitionNotAllowed(trigger_data.event, state)

        return result if executed else None

    async def _activate(self, trigger_data: TriggerData, transition: "Transition"):
        event_data = EventData(trigger_data=trigger_data, transition=transition)
        args, kwargs = event_data.args, event_data.extended_kwargs

        await self.sm._callbacks.async_call(transition.validators.key, *args, **kwargs)
        if not await self.sm._callbacks.async_all(transition.cond.key, *args, **kwargs):
            return False, None

        source = transition.source
        target = transition.target

        result = await self.sm._callbacks.async_call(transition.before.key, *args, **kwargs)
        if source is not None and not transition.internal:
            await self.sm._callbacks.async_call(source.exit.key, *args, **kwargs)

        result += await self.sm._callbacks.async_call(transition.on.key, *args, **kwargs)

        self.sm.current_state = target
        event_data.state = target
        kwargs["state"] = target

        if not transition.internal:
            await self.sm._callbacks.async_call(target.enter.key, *args, **kwargs)
        await self.sm._callbacks.async_call(transition.after.key, *args, **kwargs)

        if len(result) == 0:
            result = None
        elif len(result) == 1:
            result = result[0]

        return True, result



from __future__ import absolute_import

from .snappy import (
    compress,
    decompress,
    uncompress,
    stream_compress,
    stream_decompress,
    StreamCompressor,
    StreamDecompressor,
    UncompressError,
    HadoopStreamCompressor,
    HadoopStreamDecompressor,
    isValidCompressed,
)

__version__ = '0.7.3'



"""Scheduler for Python functions.

.. note::
    This is used for the thread-based worker only,
    not for amqp/redis/sqs/qpid where :mod:`kombu.asynchronous.timer` is used.
"""
import os
import sys
import threading
from itertools import count
from threading import TIMEOUT_MAX as THREAD_TIMEOUT_MAX
from time import sleep
from typing import Any, Callable, Iterator, Optional, Tuple

from kombu.asynchronous.timer import Entry
from kombu.asynchronous.timer import Timer as Schedule
from kombu.asynchronous.timer import logger, to_timestamp

TIMER_DEBUG = os.environ.get('TIMER_DEBUG')

__all__ = ('Entry', 'Schedule', 'Timer', 'to_timestamp')


class Timer(threading.Thread):
    """Timer thread.

    Note:
        This is only used for transports not supporting AsyncIO.
    """

    Entry = Entry
    Schedule = Schedule

    running: bool = False
    on_tick: Optional[Callable[[float], None]] = None

    _timer_count: count = count(1)

    if TIMER_DEBUG:  # pragma: no cover
        def start(self, *args: Any, **kwargs: Any) -> None:
            import traceback
            print('- Timer starting')
            traceback.print_stack()
            super().start(*args, **kwargs)

    def __init__(self, schedule: Optional[Schedule] = None,
                 on_error: Optional[Callable[[Exception], None]] = None,
                 on_tick: Optional[Callable[[float], None]] = None,
                 on_start: Optional[Callable[['Timer'], None]] = None,
                 max_interval: Optional[float] = None, **kwargs: Any) -> None:
        self.schedule = schedule or self.Schedule(on_error=on_error,
                                                  max_interval=max_interval)
        self.on_start = on_start
        self.on_tick = on_tick or self.on_tick
        super().__init__()
        # `_is_stopped` is likely to be an attribute on `Thread` objects so we
        # double underscore these names to avoid shadowing anything and
        # potentially getting confused by the superclass turning these into
        # something other than an `Event` instance (e.g. a `bool`)
        self.__is_shutdown = threading.Event()
        self.__is_stopped = threading.Event()
        self.mutex = threading.Lock()
        self.not_empty = threading.Condition(self.mutex)
        self.daemon = True
        self.name = f'Timer-{next(self._timer_count)}'

    def _next_entry(self) -> Optional[float]:
        with self.not_empty:
            delay: Optional[float]
            entry: Optional[Entry]
            delay, entry = next(self.scheduler)
            if entry is None:
                if delay is None:
                    self.not_empty.wait(1.0)
                return delay
        return self.schedule.apply_entry(entry)
    __next__ = next = _next_entry  # for 2to3

    def run(self) -> None:
        try:
            self.running = True
            self.scheduler: Iterator[Tuple[Optional[float], Optional[Entry]]] = iter(self.schedule)

            while not self.__is_shutdown.is_set():
                delay = self._next_entry()
                if delay:
                    if self.on_tick:
                        self.on_tick(delay)
                    if sleep is None:  # pragma: no cover
                        break
                    sleep(delay)
            try:
                self.__is_stopped.set()
            except TypeError:  # pragma: no cover
                # we lost the race at interpreter shutdown,
                # so gc collected built-in modules.
                pass
        except Exception as exc:
            logger.error('Thread Timer crashed: %r', exc, exc_info=True)
            sys.stderr.flush()
            os._exit(1)

    def stop(self) -> None:
        self.__is_shutdown.set()
        if self.running:
            self.__is_stopped.wait()
            self.join(THREAD_TIMEOUT_MAX)
            self.running = False

    def ensure_started(self) -> None:
        if not self.running and not self.is_alive():
            if self.on_start:
                self.on_start(self)
            self.start()

    def _do_enter(self, meth: str, *args: Any, **kwargs: Any) -> Entry:
        self.ensure_started()
        with self.mutex:
            entry = getattr(self.schedule, meth)(*args, **kwargs)
            self.not_empty.notify()
            return entry

    def enter(self, entry: Entry, eta: float, priority: Optional[int] = None) -> Entry:
        return self._do_enter('enter_at', entry, eta, priority=priority)

    def call_at(self, *args: Any, **kwargs: Any) -> Entry:
        return self._do_enter('call_at', *args, **kwargs)

    def enter_after(self, *args: Any, **kwargs: Any) -> Entry:
        return self._do_enter('enter_after', *args, **kwargs)

    def call_after(self, *args: Any, **kwargs: Any) -> Entry:
        return self._do_enter('call_after', *args, **kwargs)

    def call_repeatedly(self, *args: Any, **kwargs: Any) -> Entry:
        return self._do_enter('call_repeatedly', *args, **kwargs)

    def exit_after(self, secs: float, priority: int = 10) -> None:
        self.call_after(secs, sys.exit, priority)

    def cancel(self, tref: Entry) -> None:
        tref.cancel()

    def clear(self) -> None:
        self.schedule.clear()

    def empty(self) -> bool:
        return not len(self)

    def __len__(self) -> int:
        return len(self.schedule)

    def __bool__(self) -> bool:
        """``bool(timer)``."""
        return True
    __nonzero__ = __bool__

    @property
    def queue(self) -> list:
        return self.schedule.queue



import json
import uuid
from typing import Optional

import ulid
from loguru import logger
from statemachine.exceptions import TransitionNotAllowed

from genie_flow.celery import CeleryManager
from genie_flow.celery.transition import TransitionManager
from genie_flow.environment import GenieEnvironment
from genie_flow.genie import GenieModel
from genie_flow.model.persistence import PersistenceLevel, Persistence
from genie_flow.model.types import ModelKeyRegistryType
from genie_flow.model.api import AIResponse, EventInput, AIStatusResponse, AIProgressResponse
from genie_flow.mongo import retrieve_user_sessions_mongo
from genie_flow.session_lock import SessionLockManager
from genie_flow.model.user import User

class SessionManager:
    """
    A `SessionManager` instance deals with the lifetime events against the state machine of a
    session. From conception (through a `start_session` call), to handling events being sent
    to the state machine.
    """

    def __init__(
        self,
        session_lock_manager: SessionLockManager,
        model_key_registry: ModelKeyRegistryType,
        genie_environment: GenieEnvironment,
        celery_manager: CeleryManager,
    ):
        self.session_lock_manager = session_lock_manager
        self.model_key_registry = model_key_registry
        self.genie_environment = genie_environment
        self.celery_manager = celery_manager

    def _create_new_model(
            self,
            model_key: str,
            persistence: Persistence,
            user_info: Optional[User]=None,
    ):
        """
        Create a new model instance for a given model key with specified persistence level.
    
        This method creates a new session ID using ULID and instantiates a new model instance
        of the class registered under the given model key. The model is initialized with the
        provided persistence level and optional user information.
    
        :param model_key: The key under which the model class is registered
        :param persistence: Persistence configuration specifying how the model should be stored
        :param user_info: Optional User instance containing information about the current user
        :return: A new instance of the model class registered under the given model key
        """

        session_id = str(ulid.new().uuid)
        logger.info(
            "Creating new session {session_id} for model {model_key} "
            "and persistence level {persistence_level}",
            session_id=session_id,
            model_key=model_key,
            persistence_level=persistence.level.name,
        )

        model_class = self.model_key_registry[model_key]
        model = model_class(session_id=session_id)

        model.secondary_storage["persistence"] = persistence
        if user_info is not None:
            model.secondary_storage["user_info"] = user_info

        return model

    def get_user_sessions(self, user_info: User):
        return retrieve_user_sessions_mongo(user_info)

    def create_new_session(self, model_key: str, user_info: Optional[User]=None) -> AIResponse:
        """
        Create a new session. This method creates a new session id (ULID), creates a model
        instance of the given model class, initiates a state machine for that model instance
        and finally persists the model to Redis.

        The state machine is initiated with the `new_session` flag true, forcing it to place
        the state machine into the initial state and setting the appropriate values of the new
        model instance.

        The method then returns the appropriate `AIResponse` object with the (initial) response
        and the actions that the state machine can take from this initial state.

        :param model_key: the key under which the model class is registered
        :param user_info: optional User instance containing information about the current user
        :return: an instance of `AIResponse` with the appropriate values
        """
        model = self._create_new_model(
            model_key=model_key,
            persistence=Persistence(level=PersistenceLevel.LONG_TERM_PERSISTENCE),
            user_info=user_info,
        )
        state_machine = model.get_state_machine_class()(model)

        initial_prompt = self.genie_environment.render_template(
            state_machine.get_template_for_state(state_machine.current_state),
            model.render_data,
        )
        model.add_dialogue_element("assistant", initial_prompt)
        self.session_lock_manager.store_model(model)

        response = model.current_response.actor_text

        return AIResponse(
            session_id=model.session_id,
            response=response,
            next_actions=state_machine.current_state.transitions.unique_events,
        )

    def start_ephemeral_session(
            self,
            model_key: str,
            event: str,
            event_input: str,
            user_info: Optional[User]=None,
    ) -> AIResponse:
        """
        Create a new ephemeral session. This method creates a new session id (ULID),
        creates a model and initiates a state machine for that model instance. It then
        immediately sends the event to the state machine and returns the appropriate
        nex actions that can be taken from the current state.
    
        :param model_key: the key under which the model class is registered
        :param event: the event to send to the state machine
        :param event_input: the input to the event
        :param user_info: optional User instance containing information about the current user
        :return: an instance of `AIResponse` with the appropriate values
        """
        model = self._create_new_model(
            model_key=model_key,
            persistence=Persistence(level=PersistenceLevel.EPHEMERAL),
            user_info=user_info,
        )

        self.session_lock_manager.store_model(model)

        return self.process_event(
            model_key,
            EventInput(
                session_id=model.session_id,
                event=event,
                event_input=event_input,
            )
        )

    def _handle_poll(self, model: GenieModel) -> AIResponse:
        """
        This method handles polling from the client. As long as the model instance has a value
        for `running_task_id`, this method returns an AIResponse object with the only possible
        next actions to be `poll`.

        If the model instance does no longer have a running task (because that was finished)
        an AIResponse object is created with the session id, the most recently recorded actor
        text and the events that can be sent from the current state.

        :param model: the model that needs to be polled
        :return: an instance of `AIResponse` with the appropriate values
        """
        if self.session_lock_manager.progress_exists(model.session_id):
            todo, done = self.session_lock_manager.progress_status(model.session_id)
            return AIResponse(
                session_id=model.session_id,
                next_actions=["poll"],
                progress=AIProgressResponse(
                    total_number_of_subtasks=todo,
                    number_of_subtasks_executed=done,
                )
            )

        state_machine = model.get_state_machine_class()(model)
        if model.has_errors:
            return AIResponse(
                session_id=model.session_id,
                error=model.task_error,
                next_actions=state_machine.current_state.transitions.unique_events,
            )
        try:
            actor_response = state_machine.model.current_response.actor_text
        except AttributeError:
            logger.warning(
                "There is no recorded actor response for session {session_id}",
            )
            actor_response = ""

        return AIResponse(
            session_id=model.session_id,
            response=actor_response,
            next_actions=state_machine.current_state.transitions.unique_events,
        )

    def _handle_event(self, event: EventInput, model: GenieModel) -> AIResponse:
        """
        This method handels events from the client. It creates the state machine instance for the
        given object and sends the event to it. It then stores the model instance back into Redis.

        If the state machine, after processing the given event, has a currently running task,
        this method returns an AIResponse object with the only next actions to be `poll`.

        If the processing of the event by the state machine has not resulted in a task, this method
        returns an AIResponse object with the most recently recorded actor text and the events that
        can be sent from the current state.

        Session locking, saving and storing of the model object needs to happen outside of
        this method.

        :param event: the event to process
        :param model: the model to process the event against
        :return: an instance of `AIResponse` with the appropriate values
        """
        state_machine = model.get_state_machine_class()(model)
        state_machine.add_listener(TransitionManager(self.celery_manager))
        state_machine.send(event.event, event.event_input)

        if self.session_lock_manager.progress_exists(model.session_id):
            return AIResponse(session_id=event.session_id, next_actions=["poll"])

        return AIResponse(
            session_id=event.session_id,
            response=state_machine.model.current_response.actor_text,
            next_actions=state_machine.current_state.transitions.unique_events,
        )

    def process_event(self, model_key: str, event: EventInput) -> AIResponse:
        """
        Process incoming events. Claims a lock to the model instance that the event refers to
        and checks the event. If the event is a `poll` event, handling is performed by the
        `_handle_poll` method. If not, this method returns the result of processing the event.

        :param model_key: the key under which the model class is registered
        :param event: the event to process
        :return: an instance of `AIResponse` with the appropriate values
        """
        model_class = self.model_key_registry[model_key]
        with self.session_lock_manager.get_locked_model(event.session_id, model_class) as model:
            if event.event == "poll":
                return self._handle_poll(model)

            try:
                return self._handle_event(event, model)
            except TransitionNotAllowed:
                state_machine = model.get_state_machine_class()(model)
                return AIResponse(
                    session_id=event.session_id,
                    error=json.dumps(
                        dict(
                            session_id=model.session_id,
                            current_state=dict(
                                id=state_machine.current_state.id,
                                name=state_machine.current_state.name,
                            ),
                            possible_events=state_machine.current_state.transitions.unique_events,
                            received_event=event.event,
                        )
                    )
                )

    def get_task_state(self, model_key: str, session_id: str) -> AIStatusResponse:
        """
        Retrieves an instance of the model object and returns if that object has any running
        tasks against it. It obtains a lock on the given session id to ensure consistency of
        the model values.

        The `AIStatusResponse` that this method returns indicates if the task is currently running,
        or, if it is no longer running, what the possible next actions are.

        :param model_key: the key under which the model class is registered
        :param session_id: the id of the session that the model instance belongs to
        :return: an instance of `AIStatusResponse`, indicating if the task is ready and what
        possible next actions can be sent in the current state of the model.
        """
        if self.session_lock_manager.progress_exists(session_id):
            return AIStatusResponse(
                session_id=session_id,
                ready=False,
            )

        model_class = self.model_key_registry[model_key]
        model = self.session_lock_manager.get_model(session_id, model_class)
        state_machine = model.get_state_machine_class()(model=model)
        return AIStatusResponse(
            session_id=session_id,
            ready=True,
            next_actions=state_machine.current_state.transitions.unique_events,
        )

    def get_model(self, model_key: str, session_id: str) -> GenieModel:
        """
        Retrieve the entire model instance that belongs to the given session id. Obtains a lock
        on the session to ensure consistency of the model values.

        :param model_key: the key under which the model class is registered
        :param session_id: the session id to retrieve the model instance for
        :return: the model instance that belongs to the given session id
        """
        model_class = self.model_key_registry[model_key]
        return self.session_lock_manager.get_model(session_id, model_class)



"""SQLAlchemy result store backend."""
import logging
from contextlib import contextmanager

from vine.utils import wraps

from celery import states
from celery.backends.base import BaseBackend
from celery.exceptions import ImproperlyConfigured
from celery.utils.time import maybe_timedelta

from .models import Task, TaskExtended, TaskSet
from .session import SessionManager

try:
    from sqlalchemy.exc import DatabaseError, InvalidRequestError
    from sqlalchemy.orm.exc import StaleDataError
except ImportError:
    raise ImproperlyConfigured(
        'The database result backend requires SQLAlchemy to be installed.'
        'See https://pypi.org/project/SQLAlchemy/')

logger = logging.getLogger(__name__)

__all__ = ('DatabaseBackend',)


@contextmanager
def session_cleanup(session):
    try:
        yield
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()


def retry(fun):

    @wraps(fun)
    def _inner(*args, **kwargs):
        max_retries = kwargs.pop('max_retries', 3)

        for retries in range(max_retries):
            try:
                return fun(*args, **kwargs)
            except (DatabaseError, InvalidRequestError, StaleDataError):
                logger.warning(
                    'Failed operation %s.  Retrying %s more times.',
                    fun.__name__, max_retries - retries - 1,
                    exc_info=True)
                if retries + 1 >= max_retries:
                    raise

    return _inner


class DatabaseBackend(BaseBackend):
    """The database result backend."""

    # ResultSet.iterate should sleep this much between each pool,
    # to not bombard the database with queries.
    subpolling_interval = 0.5

    task_cls = Task
    taskset_cls = TaskSet

    def __init__(self, dburi=None, engine_options=None, url=None, **kwargs):
        # The `url` argument was added later and is used by
        # the app to set backend by url (celery.app.backends.by_url)
        super().__init__(expires_type=maybe_timedelta,
                         url=url, **kwargs)
        conf = self.app.conf

        if self.extended_result:
            self.task_cls = TaskExtended

        self.url = url or dburi or conf.database_url
        self.engine_options = dict(
            engine_options or {},
            **conf.database_engine_options or {})
        self.short_lived_sessions = kwargs.get(
            'short_lived_sessions',
            conf.database_short_lived_sessions)

        schemas = conf.database_table_schemas or {}
        tablenames = conf.database_table_names or {}
        self.task_cls.configure(
            schema=schemas.get('task'),
            name=tablenames.get('task'))
        self.taskset_cls.configure(
            schema=schemas.get('group'),
            name=tablenames.get('group'))

        if not self.url:
            raise ImproperlyConfigured(
                'Missing connection string! Do you have the'
                ' database_url setting set to a real value?')

        self.session_manager = SessionManager()

        create_tables_at_setup = conf.database_create_tables_at_setup
        if create_tables_at_setup is True:
            self._create_tables()

    @property
    def extended_result(self):
        return self.app.conf.find_value_for_key('extended', 'result')

    def _create_tables(self):
        """Create the task and taskset tables."""
        self.ResultSession()

    def ResultSession(self, session_manager=None):
        if session_manager is None:
            session_manager = self.session_manager
        return session_manager.session_factory(
            dburi=self.url,
            short_lived_sessions=self.short_lived_sessions,
            **self.engine_options)

    @retry
    def _store_result(self, task_id, result, state, traceback=None,
                      request=None, **kwargs):
        """Store return value and state of an executed task."""
        session = self.ResultSession()
        with session_cleanup(session):
            task = list(session.query(self.task_cls).filter(self.task_cls.task_id == task_id))
            task = task and task[0]
            if not task:
                task = self.task_cls(task_id)
                task.task_id = task_id
                session.add(task)
                session.flush()

            self._update_result(task, result, state, traceback=traceback, request=request)
            session.commit()

    def _update_result(self, task, result, state, traceback=None,
                       request=None):

        meta = self._get_result_meta(result=result, state=state,
                                     traceback=traceback, request=request,
                                     format_date=False, encode=True)

        # Exclude the primary key id and task_id columns
        # as we should not set it None
        columns = [column.name for column in self.task_cls.__table__.columns
                   if column.name not in {'id', 'task_id'}]

        # Iterate through the columns name of the table
        # to set the value from meta.
        # If the value is not present in meta, set None
        for column in columns:
            value = meta.get(column)
            setattr(task, column, value)

    @retry
    def _get_task_meta_for(self, task_id):
        """Get task meta-data for a task by id."""
        session = self.ResultSession()
        with session_cleanup(session):
            task = list(session.query(self.task_cls).filter(self.task_cls.task_id == task_id))
            task = task and task[0]
            if not task:
                task = self.task_cls(task_id)
                task.status = states.PENDING
                task.result = None
            data = task.to_dict()
            if data.get('args', None) is not None:
                data['args'] = self.decode(data['args'])
            if data.get('kwargs', None) is not None:
                data['kwargs'] = self.decode(data['kwargs'])
            return self.meta_from_decoded(data)

    @retry
    def _save_group(self, group_id, result):
        """Store the result of an executed group."""
        session = self.ResultSession()
        with session_cleanup(session):
            group = self.taskset_cls(group_id, result)
            session.add(group)
            session.flush()
            session.commit()
            return result

    @retry
    def _restore_group(self, group_id):
        """Get meta-data for group by id."""
        session = self.ResultSession()
        with session_cleanup(session):
            group = session.query(self.taskset_cls).filter(
                self.taskset_cls.taskset_id == group_id).first()
            if group:
                return group.to_dict()

    @retry
    def _delete_group(self, group_id):
        """Delete meta-data for group by id."""
        session = self.ResultSession()
        with session_cleanup(session):
            session.query(self.taskset_cls).filter(
                self.taskset_cls.taskset_id == group_id).delete()
            session.flush()
            session.commit()

    @retry
    def _forget(self, task_id):
        """Forget about result."""
        session = self.ResultSession()
        with session_cleanup(session):
            session.query(self.task_cls).filter(self.task_cls.task_id == task_id).delete()
            session.commit()

    def cleanup(self):
        """Delete expired meta-data."""
        session = self.ResultSession()
        expires = self.expires
        now = self.app.now()
        with session_cleanup(session):
            session.query(self.task_cls).filter(
                self.task_cls.date_done < (now - expires)).delete()
            session.query(self.taskset_cls).filter(
                self.taskset_cls.date_done < (now - expires)).delete()
            session.commit()

    def __reduce__(self, args=(), kwargs=None):
        kwargs = {} if not kwargs else kwargs
        kwargs.update(
            {'dburi': self.url,
             'expires': self.expires,
             'engine_options': self.engine_options})
        return super().__reduce__(args, kwargs)



"""AMQP Connections."""
# Copyright (C) 2007-2008 Barry Pederson <bp@barryp.org>

import logging
import socket
import uuid
import warnings
from array import array
from time import monotonic

from vine import ensure_promise

from . import __version__, sasl, spec
from .abstract_channel import AbstractChannel
from .channel import Channel
from .exceptions import (AMQPDeprecationWarning, ChannelError, ConnectionError,
                         ConnectionForced, MessageNacked, RecoverableChannelError,
                         RecoverableConnectionError, ResourceError,
                         error_for_code)
from .method_framing import frame_handler, frame_writer
from .transport import Transport

try:
    from ssl import SSLError
except ImportError:  # pragma: no cover
    class SSLError(Exception):  # noqa
        pass

W_FORCE_CONNECT = """\
The .{attr} attribute on the connection was accessed before
the connection was established.  This is supported for now, but will
be deprecated in amqp 2.2.0.

Since amqp 2.0 you have to explicitly call Connection.connect()
before using the connection.
"""

START_DEBUG_FMT = """
Start from server, version: %d.%d, properties: %s, mechanisms: %s, locales: %s
""".strip()

__all__ = ('Connection',)

AMQP_LOGGER = logging.getLogger('amqp')
AMQP_HEARTBEAT_LOGGER = logging.getLogger(
    'amqp.connection.Connection.heartbeat_tick'
)

#: Default map for :attr:`Connection.library_properties`
LIBRARY_PROPERTIES = {
    'product': 'py-amqp',
    'product_version': __version__,
}

#: Default map for :attr:`Connection.negotiate_capabilities`
NEGOTIATE_CAPABILITIES = {
    'consumer_cancel_notify': True,
    'connection.blocked': True,
    'authentication_failure_close': True,
}


class Connection(AbstractChannel):
    """AMQP Connection.

    The connection class provides methods for a client to establish a
    network connection to a server, and for both peers to operate the
    connection thereafter.

    GRAMMAR::

        connection          = open-connection *use-connection close-connection
        open-connection     = C:protocol-header
                              S:START C:START-OK
                              *challenge
                              S:TUNE C:TUNE-OK
                              C:OPEN S:OPEN-OK
        challenge           = S:SECURE C:SECURE-OK
        use-connection      = *channel
        close-connection    = C:CLOSE S:CLOSE-OK
                            / S:CLOSE C:CLOSE-OK
    Create a connection to the specified host, which should be
    a 'host[:port]', such as 'localhost', or '1.2.3.4:5672'
    (defaults to 'localhost', if a port is not specified then
    5672 is used)

    Authentication can be controlled by passing one or more
    `amqp.sasl.SASL` instances as the `authentication` parameter, or
    setting the `login_method` string to one of the supported methods:
    'GSSAPI', 'EXTERNAL', 'AMQPLAIN', or 'PLAIN'.
    Otherwise authentication will be performed using any supported method
    preferred by the server. Userid and passwords apply to AMQPLAIN and
    PLAIN authentication, whereas on GSSAPI only userid will be used as the
    client name. For EXTERNAL authentication both userid and password are
    ignored.

    The 'ssl' parameter may be simply True/False, or
    a dictionary of options to pass to :class:`ssl.SSLContext` such as
    requiring certain certificates. For details, refer ``ssl`` parameter of
    :class:`~amqp.transport.SSLTransport`.

    The "socket_settings" parameter is a dictionary defining tcp
    settings which will be applied as socket options.

    When "confirm_publish" is set to True, the channel is put to
    confirm mode. In this mode, each published message is
    confirmed using Publisher confirms RabbitMQ extension.
    """

    Channel = Channel

    #: Mapping of protocol extensions to enable.
    #: The server will report these in server_properties[capabilities],
    #: and if a key in this map is present the client will tell the
    #: server to either enable or disable the capability depending
    #: on the value set in this map.
    #: For example with:
    #:     negotiate_capabilities = {
    #:         'consumer_cancel_notify': True,
    #:     }
    #: The client will enable this capability if the server reports
    #: support for it, but if the value is False the client will
    #: disable the capability.
    negotiate_capabilities = NEGOTIATE_CAPABILITIES

    #: These are sent to the server to announce what features
    #: we support, type of client etc.
    library_properties = LIBRARY_PROPERTIES

    #: Final heartbeat interval value (in float seconds) after negotiation
    heartbeat = None

    #: Original heartbeat interval value proposed by client.
    client_heartbeat = None

    #: Original heartbeat interval proposed by server.
    server_heartbeat = None

    #: Time of last heartbeat sent (in monotonic time, if available).
    last_heartbeat_sent = 0

    #: Time of last heartbeat received (in monotonic time, if available).
    last_heartbeat_received = 0

    #: Number of successful writes to socket.
    bytes_sent = 0

    #: Number of successful reads from socket.
    bytes_recv = 0

    #: Number of bytes sent to socket at the last heartbeat check.
    prev_sent = None

    #: Number of bytes received from socket at the last heartbeat check.
    prev_recv = None

    _METHODS = {
        spec.method(spec.Connection.Start, 'ooFSS'),
        spec.method(spec.Connection.OpenOk),
        spec.method(spec.Connection.Secure, 's'),
        spec.method(spec.Connection.Tune, 'BlB'),
        spec.method(spec.Connection.Close, 'BsBB'),
        spec.method(spec.Connection.Blocked),
        spec.method(spec.Connection.Unblocked),
        spec.method(spec.Connection.CloseOk),
    }
    _METHODS = {m.method_sig: m for m in _METHODS}

    _ALLOWED_METHODS_WHEN_CLOSING = (
        spec.Connection.Close, spec.Connection.CloseOk
    )

    connection_errors = (
        ConnectionError,
        socket.error,
        IOError,
        OSError,
    )
    channel_errors = (ChannelError,)
    recoverable_connection_errors = (
        RecoverableConnectionError,
        MessageNacked,
        socket.error,
        IOError,
        OSError,
    )
    recoverable_channel_errors = (
        RecoverableChannelError,
    )

    def __init__(self, host='localhost:5672', userid='guest', password='guest',
                 login_method=None, login_response=None,
                 authentication=(),
                 virtual_host='/', locale='en_US', client_properties=None,
                 ssl=False, connect_timeout=None, channel_max=None,
                 frame_max=None, heartbeat=0, on_open=None, on_blocked=None,
                 on_unblocked=None, confirm_publish=False,
                 on_tune_ok=None, read_timeout=None, write_timeout=None,
                 socket_settings=None, frame_handler=frame_handler,
                 frame_writer=frame_writer, **kwargs):
        self._connection_id = uuid.uuid4().hex
        channel_max = channel_max or 65535
        frame_max = frame_max or 131072
        if authentication:
            if isinstance(authentication, sasl.SASL):
                authentication = (authentication,)
            self.authentication = authentication
        elif login_method is not None:
            if login_method == 'GSSAPI':
                auth = sasl.GSSAPI(userid)
            elif login_method == 'EXTERNAL':
                auth = sasl.EXTERNAL()
            elif login_method == 'AMQPLAIN':
                if userid is None or password is None:
                    raise ValueError(
                        "Must supply authentication or userid/password")
                auth = sasl.AMQPLAIN(userid, password)
            elif login_method == 'PLAIN':
                if userid is None or password is None:
                    raise ValueError(
                        "Must supply authentication or userid/password")
                auth = sasl.PLAIN(userid, password)
            elif login_response is not None:
                auth = sasl.RAW(login_method, login_response)
            else:
                raise ValueError("Invalid login method", login_method)
            self.authentication = (auth,)
        else:
            self.authentication = (sasl.GSSAPI(userid, fail_soft=True),
                                   sasl.EXTERNAL(),
                                   sasl.AMQPLAIN(userid, password),
                                   sasl.PLAIN(userid, password))

        self.client_properties = dict(
            self.library_properties, **client_properties or {}
        )
        self.locale = locale
        self.host = host
        self.virtual_host = virtual_host
        self.on_tune_ok = ensure_promise(on_tune_ok)

        self.frame_handler_cls = frame_handler
        self.frame_writer_cls = frame_writer

        self._handshake_complete = False

        self.channels = {}
        # The connection object itself is treated as channel 0
        super().__init__(self, 0)

        self._frame_writer = None
        self._on_inbound_frame = None
        self._transport = None

        # Properties set in the Tune method
        self.channel_max = channel_max
        self.frame_max = frame_max
        self.client_heartbeat = heartbeat

        self.confirm_publish = confirm_publish
        self.ssl = ssl
        self.read_timeout = read_timeout
        self.write_timeout = write_timeout
        self.socket_settings = socket_settings

        # Callbacks
        self.on_blocked = on_blocked
        self.on_unblocked = on_unblocked
        self.on_open = ensure_promise(on_open)

        self._used_channel_ids = array('H')

        # Properties set in the Start method
        self.version_major = 0
        self.version_minor = 0
        self.server_properties = {}
        self.mechanisms = []
        self.locales = []

        self.connect_timeout = connect_timeout

    def __repr__(self):
        if self._transport:
            return f'<AMQP Connection: {self.host}/{self.virtual_host} '\
                   f'using {self._transport} at {id(self):#x}>'
        else:
            return f'<AMQP Connection: {self.host}/{self.virtual_host} '\
                   f'(disconnected) at {id(self):#x}>'

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, *eargs):
        self.close()

    def then(self, on_success, on_error=None):
        return self.on_open.then(on_success, on_error)

    def _setup_listeners(self):
        self._callbacks.update({
            spec.Connection.Start: self._on_start,
            spec.Connection.OpenOk: self._on_open_ok,
            spec.Connection.Secure: self._on_secure,
            spec.Connection.Tune: self._on_tune,
            spec.Connection.Close: self._on_close,
            spec.Connection.Blocked: self._on_blocked,
            spec.Connection.Unblocked: self._on_unblocked,
            spec.Connection.CloseOk: self._on_close_ok,
        })

    def connect(self, callback=None):
        # Let the transport.py module setup the actual
        # socket connection to the broker.
        #
        if self.connected:
            return callback() if callback else None
        try:
            self.transport = self.Transport(
                self.host, self.connect_timeout, self.ssl,
                self.read_timeout, self.write_timeout,
                socket_settings=self.socket_settings,
            )
            self.transport.connect()
            self.on_inbound_frame = self.frame_handler_cls(
                self, self.on_inbound_method)
            self.frame_writer = self.frame_writer_cls(self, self.transport)

            while not self._handshake_complete:
                self.drain_events(timeout=self.connect_timeout)

        except (OSError, SSLError):
            self.collect()
            raise

    def _warn_force_connect(self, attr):
        warnings.warn(AMQPDeprecationWarning(
            W_FORCE_CONNECT.format(attr=attr)))

    @property
    def transport(self):
        if self._transport is None:
            self._warn_force_connect('transport')
            self.connect()
        return self._transport

    @transport.setter
    def transport(self, transport):
        self._transport = transport

    @property
    def on_inbound_frame(self):
        if self._on_inbound_frame is None:
            self._warn_force_connect('on_inbound_frame')
            self.connect()
        return self._on_inbound_frame

    @on_inbound_frame.setter
    def on_inbound_frame(self, on_inbound_frame):
        self._on_inbound_frame = on_inbound_frame

    @property
    def frame_writer(self):
        if self._frame_writer is None:
            self._warn_force_connect('frame_writer')
            self.connect()
        return self._frame_writer

    @frame_writer.setter
    def frame_writer(self, frame_writer):
        self._frame_writer = frame_writer

    def _on_start(self, version_major, version_minor, server_properties,
                  mechanisms, locales, argsig='FsSs'):
        client_properties = self.client_properties
        self.version_major = version_major
        self.version_minor = version_minor
        self.server_properties = server_properties
        if isinstance(mechanisms, str):
            mechanisms = mechanisms.encode('utf-8')
        self.mechanisms = mechanisms.split(b' ')
        self.locales = locales.split(' ')
        AMQP_LOGGER.debug(
            START_DEBUG_FMT,
            self.version_major, self.version_minor,
            self.server_properties, self.mechanisms, self.locales,
        )

        # Negotiate protocol extensions (capabilities)
        scap = server_properties.get('capabilities') or {}
        cap = client_properties.setdefault('capabilities', {})
        cap.update({
            wanted_cap: enable_cap
            for wanted_cap, enable_cap in self.negotiate_capabilities.items()
            if scap.get(wanted_cap)
        })
        if not cap:
            # no capabilities, server may not react well to having
            # this key present in client_properties, so we remove it.
            client_properties.pop('capabilities', None)

        for authentication in self.authentication:
            if authentication.mechanism in self.mechanisms:
                login_response = authentication.start(self)
                if login_response is not NotImplemented:
                    break
        else:
            raise ConnectionError(
                "Couldn't find appropriate auth mechanism "
                "(can offer: {}; available: {})".format(
                    b", ".join(m.mechanism
                               for m in self.authentication
                               if m.mechanism).decode(),
                    b", ".join(self.mechanisms).decode()))

        self.send_method(
            spec.Connection.StartOk, argsig,
            (client_properties, authentication.mechanism,
             login_response, self.locale),
        )

    def _on_secure(self, challenge):
        pass

    def _on_tune(self, channel_max, frame_max, server_heartbeat, argsig='BlB'):
        client_heartbeat = self.client_heartbeat or 0
        self.channel_max = channel_max or self.channel_max
        self.frame_max = frame_max or self.frame_max
        self.server_heartbeat = server_heartbeat or 0

        # negotiate the heartbeat interval to the smaller of the
        # specified values
        if self.server_heartbeat == 0 or client_heartbeat == 0:
            self.heartbeat = max(self.server_heartbeat, client_heartbeat)
        else:
            self.heartbeat = min(self.server_heartbeat, client_heartbeat)

        # Ignore server heartbeat if client_heartbeat is disabled
        if not self.client_heartbeat:
            self.heartbeat = 0

        self.send_method(
            spec.Connection.TuneOk, argsig,
            (self.channel_max, self.frame_max, self.heartbeat),
            callback=self._on_tune_sent,
        )

    def _on_tune_sent(self, argsig='ssb'):
        self.send_method(
            spec.Connection.Open, argsig, (self.virtual_host, '', False),
        )

    def _on_open_ok(self):
        self._handshake_complete = True
        self.on_open(self)

    def Transport(self, host, connect_timeout,
                  ssl=False, read_timeout=None, write_timeout=None,
                  socket_settings=None, **kwargs):
        return Transport(
            host, connect_timeout=connect_timeout, ssl=ssl,
            read_timeout=read_timeout, write_timeout=write_timeout,
            socket_settings=socket_settings, **kwargs)

    @property
    def connected(self):
        return self._transport and self._transport.connected

    def collect(self):
        if self._transport:
            self._transport.close()

        if self.channels:
            # Copy all the channels except self since the channels
            # dictionary changes during the collection process.
            channels = [
                ch for ch in self.channels.values()
                if ch is not self
            ]

            for ch in channels:
                ch.collect()
        self._transport = self.connection = self.channels = None

    def _get_free_channel_id(self):
        # Cast to a set for fast lookups, and keep stored as an array for lower memory usage.
        used_channel_ids = set(self._used_channel_ids)

        for channel_id in range(1, self.channel_max + 1):
            if channel_id not in used_channel_ids:
                self._used_channel_ids.append(channel_id)
                return channel_id

        raise ResourceError(
            'No free channel ids, current={}, channel_max={}'.format(
                len(self.channels), self.channel_max), spec.Channel.Open)

    def _claim_channel_id(self, channel_id):
        if channel_id in self._used_channel_ids:
            raise ConnectionError(f'Channel {channel_id!r} already open')
        else:
            self._used_channel_ids.append(channel_id)
            return channel_id

    def channel(self, channel_id=None, callback=None):
        """Create new channel.

        Fetch a Channel object identified by the numeric channel_id, or
        create that object if it doesn't already exist.
        """
        if self.channels is None:
            raise RecoverableConnectionError('Connection already closed.')

        try:
            return self.channels[channel_id]
        except KeyError:
            channel = self.Channel(self, channel_id, on_open=callback)
            channel.open()
            return channel

    def is_alive(self):
        raise NotImplementedError('Use AMQP heartbeats')

    def drain_events(self, timeout=None):
        # read until message is ready
        while not self.blocking_read(timeout):
            pass

    def blocking_read(self, timeout=None):
        with self.transport.having_timeout(timeout):
            frame = self.transport.read_frame()
        return self.on_inbound_frame(frame)

    def on_inbound_method(self, channel_id, method_sig, payload, content):
        if self.channels is None:
            raise RecoverableConnectionError('Connection already closed')

        return self.channels[channel_id].dispatch_method(
            method_sig, payload, content,
        )

    def close(self, reply_code=0, reply_text='', method_sig=(0, 0),
              argsig='BsBB'):
        """Request a connection close.

        This method indicates that the sender wants to close the
        connection. This may be due to internal conditions (e.g. a
        forced shut-down) or due to an error handling a specific
        method, i.e. an exception.  When a close is due to an
        exception, the sender provides the class and method id of the
        method which caused the exception.

        RULE:

            After sending this method any received method except the
            Close-OK method MUST be discarded.

        RULE:

            The peer sending this method MAY use a counter or timeout
            to detect failure of the other peer to respond correctly
            with the Close-OK method.

        RULE:

            When a server receives the Close method from a client it
            MUST delete all server-side resources associated with the
            client's context.  A client CANNOT reconnect to a context
            after sending or receiving a Close method.

        PARAMETERS:
            reply_code: short

                The reply code. The AMQ reply codes are defined in AMQ
                RFC 011.

            reply_text: shortstr

                The localised reply text.  This text can be logged as an
                aid to resolving issues.

            class_id: short

                failing method class

                When the close is provoked by a method exception, this
                is the class of the method.

            method_id: short

                failing method ID

                When the close is provoked by a method exception, this
                is the ID of the method.
        """
        if self._transport is None:
            # already closed
            return

        try:
            self.is_closing = True
            return self.send_method(
                spec.Connection.Close, argsig,
                (reply_code, reply_text, method_sig[0], method_sig[1]),
                wait=spec.Connection.CloseOk,
            )
        except (OSError, SSLError):
            # close connection
            self.collect()
            raise
        finally:
            self.is_closing = False

    def _on_close(self, reply_code, reply_text, class_id, method_id):
        """Request a connection close.

        This method indicates that the sender wants to close the
        connection. This may be due to internal conditions (e.g. a
        forced shut-down) or due to an error handling a specific
        method, i.e. an exception.  When a close is due to an
        exception, the sender provides the class and method id of the
        method which caused the exception.

        RULE:

            After sending this method any received method except the
            Close-OK method MUST be discarded.

        RULE:

            The peer sending this method MAY use a counter or timeout
            to detect failure of the other peer to respond correctly
            with the Close-OK method.

        RULE:

            When a server receives the Close method from a client it
            MUST delete all server-side resources associated with the
            client's context.  A client CANNOT reconnect to a context
            after sending or receiving a Close method.

        PARAMETERS:
            reply_code: short

                The reply code. The AMQ reply codes are defined in AMQ
                RFC 011.

            reply_text: shortstr

                The localised reply text.  This text can be logged as an
                aid to resolving issues.

            class_id: short

                failing method class

                When the close is provoked by a method exception, this
                is the class of the method.

            method_id: short

                failing method ID

                When the close is provoked by a method exception, this
                is the ID of the method.
        """
        self._x_close_ok()
        raise error_for_code(reply_code, reply_text,
                             (class_id, method_id), ConnectionError)

    def _x_close_ok(self):
        """Confirm a connection close.

        This method confirms a Connection.Close method and tells the
        recipient that it is safe to release resources for the
        connection and close the socket.

        RULE:
            A peer that detects a socket closure without having
            received a Close-Ok handshake method SHOULD log the error.
        """
        self.send_method(spec.Connection.CloseOk, callback=self._on_close_ok)

    def _on_close_ok(self):
        """Confirm a connection close.

        This method confirms a Connection.Close method and tells the
        recipient that it is safe to release resources for the
        connection and close the socket.

        RULE:

            A peer that detects a socket closure without having
            received a Close-Ok handshake method SHOULD log the error.
        """
        self.collect()

    def _on_blocked(self):
        """Callback called when connection blocked.

        Notes:
            This is an RabbitMQ Extension.
        """
        reason = 'connection blocked, see broker logs'
        if self.on_blocked:
            return self.on_blocked(reason)

    def _on_unblocked(self):
        if self.on_unblocked:
            return self.on_unblocked()

    def send_heartbeat(self):
        self.frame_writer(8, 0, None, None, None)

    def heartbeat_tick(self, rate=2):
        """Send heartbeat packets if necessary.

        Raises:
            ~amqp.exceptions.ConnectionForvced: if none have been
                received recently.

        Note:
            This should be called frequently, on the order of
            once per second.

        Keyword Arguments:
            rate (int): Number of heartbeat frames to send during the heartbeat
                        timeout
        """
        AMQP_HEARTBEAT_LOGGER.debug('heartbeat_tick : for connection %s',
                                    self._connection_id)
        if not self.heartbeat:
            return

        # If rate is wrong, let's use 2 as default
        if rate <= 0:
            rate = 2

        # treat actual data exchange in either direction as a heartbeat
        sent_now = self.bytes_sent
        recv_now = self.bytes_recv
        if self.prev_sent is None or self.prev_sent != sent_now:
            self.last_heartbeat_sent = monotonic()
        if self.prev_recv is None or self.prev_recv != recv_now:
            self.last_heartbeat_received = monotonic()

        now = monotonic()
        AMQP_HEARTBEAT_LOGGER.debug(
            'heartbeat_tick : Prev sent/recv: %s/%s, '
            'now - %s/%s, monotonic - %s, '
            'last_heartbeat_sent - %s, heartbeat int. - %s '
            'for connection %s',
            self.prev_sent, self.prev_recv,
            sent_now, recv_now, now,
            self.last_heartbeat_sent,
            self.heartbeat,
            self._connection_id,
        )

        self.prev_sent, self.prev_recv = sent_now, recv_now

        # send a heartbeat if it's time to do so
        if now > self.last_heartbeat_sent + self.heartbeat / rate:
            AMQP_HEARTBEAT_LOGGER.debug(
                'heartbeat_tick: sending heartbeat for connection %s',
                self._connection_id)
            self.send_heartbeat()
            self.last_heartbeat_sent = monotonic()

        # if we've missed two intervals' heartbeats, fail; this gives the
        # server enough time to send heartbeats a little late
        two_heartbeats = 2 * self.heartbeat
        two_heartbeats_interval = self.last_heartbeat_received + two_heartbeats
        heartbeats_missed = two_heartbeats_interval < monotonic()
        if self.last_heartbeat_received and heartbeats_missed:
            raise ConnectionForced('Too many heartbeats missed')

    @property
    def sock(self):
        return self.transport.sock

    @property
    def server_capabilities(self):
        return self.server_properties.get('capabilities') or {}



from loguru import logger

from pymongo import MongoClient
from dependency_injector.wiring import inject, Provide

from genie_flow.containers.perm_persistence import (
    GenieFlowPermanentPersistenceContainer,
)
from genie_flow.genie import GenieModel
from genie_flow.model.user import User


@inject
def store_session(
    session_model: GenieModel,
    mongo_store=Provide[GenieFlowPermanentPersistenceContainer.mongo_store],
    compression=Provide[
        GenieFlowPermanentPersistenceContainer.config.object_store.object_compression
    ]
    or True,
):
    col = mongo_store["genie_db"].session_collection
    if col.find_one({"session_id": session_model.session_id}) is None:
        result = col.insert_one(
            {
                "session_id": session_model.session_id,
                "model": session_model.serialize(compression),
            }
        )
    else:
        result = col.replace_one(
            {"session_id": session_model.session_id},
            {
                "session_id": session_model.session_id,
                "model": session_model.serialize(compression),
            },
            True,
        )

    logger.debug(
        "Successfully updated mongodb: {mongo_result}", mongo_result=result.acknowledged
    )


@inject
def store_user(
    user_info: User,
    session_id: str,
    mongo_store: MongoClient = Provide[
        GenieFlowPermanentPersistenceContainer.mongo_store
    ],
):
    col = mongo_store["genie_db"].user_collection
    if col.find_one({"email": user_info.email}) is None:
        logger.debug(user_info)
        result = col.insert_one(
            {
                "email": user_info.email,
                "firstname": user_info.firstname,
                "lastname": user_info.lastname,
                "custom_properties": user_info.custom_properties,
                "sessions": [session_id],
            }
        )
    else:
        logger.debug(user_info)
        result = col.update_one(
            {"email": user_info.email}, {"$addToSet": {"sessions": session_id}}, True
        )
    logger.debug(
        "Successfully updated mongodb: {mongo_result}", mongo_result=result.acknowledged
    )


@inject
def retrieve_user_sessions_mongo(
    user_info: User,
    mongo_store: MongoClient = Provide[
        GenieFlowPermanentPersistenceContainer.mongo_store
    ],
):
    col = mongo_store["genie_db"].user_collection
    user_email = user_info.email
    result = col.find_one({"email": user_email}, {"_id": 0})
    logger.debug("retrieved user sessions {result}", result=result)
    return result


@inject
def retrieve_model(
    session_id: str,
    mongo_store: MongoClient = Provide[
        GenieFlowPermanentPersistenceContainer.mongo_store
    ],
) -> dict[str, bytes] | None:
    col = mongo_store["genie_db"].session_collection
    payload = col.find_one({"session_id": session_id}, {"model": 1, "_id": 0})
    if payload is None:
        raise KeyError(f"No model with id {session_id}")
    return payload



"""Consts and function to handle target format.
ALL_SUPPORTED_FORMATS - list of supported formats
get_decompress_function - returns stream decompress function for a current
    format (specified or autodetected)
get_compress_function - returns compress function for a current format
    (specified or default)
"""
from __future__ import absolute_import

from .snappy import (
    HadoopStreamDecompressor, StreamDecompressor,
    hadoop_stream_compress, hadoop_stream_decompress, raw_stream_compress,
    raw_stream_decompress, stream_compress, stream_decompress,
    UncompressError
)


# Means format auto detection.
# For compression will be used framing format.
# In case of decompression will try to detect a format from the input stream
# header.
DEFAULT_FORMAT = "auto"

ALL_SUPPORTED_FORMATS = ["framing", "auto"]

_COMPRESS_METHODS = {
    "framing": stream_compress,
    "hadoop": hadoop_stream_compress,
    "raw": raw_stream_compress
}

_DECOMPRESS_METHODS = {
    "framing": stream_decompress,
    "hadoop": hadoop_stream_decompress,
    "raw": raw_stream_decompress
}

# We will use framing format as the default to compression.
# And for decompression, if it's not defined explicitly, we will try to
# guess the format from the file header.
_DEFAULT_COMPRESS_FORMAT = "framing"


def uvarint(fin):
    """Read uint64 nbumber from varint encoding in a stream"""
    result = 0
    shift = 0
    while True:
        byte = fin.read(1)[0]
        result |= (byte & 0x7F) << shift
        if (byte & 0x80) == 0:
            break
        shift += 7
    return result


def check_unframed_format(fin, reset=False):
    """Can this be read using the raw codec

    This function wil return True for all snappy raw streams, but
    True does not mean that we can necessarily decode the stream.
    """
    if reset:
        fin.seek(0)
    try:
        size = uvarint(fin)
        assert size < 2**32 - 1
        next_byte = fin.read(1)[0]
        end = fin.seek(0, 2)
        assert size < end
        assert next_byte & 0b11 == 0 # must start with literal block
        return True
    except:
        return False


# The tuple contains an ordered sequence of a format checking function and
# a format-specific decompression function.
# Framing format has it's header, that may be recognized.
_DECOMPRESS_FORMAT_FUNCS = {
    "framed": stream_decompress,
    "hadoop": hadoop_stream_decompress,
    "raw": raw_stream_decompress
}


def guess_format_by_header(fin):
    """Tries to guess a compression format for the given input file by it's
    header.

    :return: format name (str), stream decompress function (callable)
    """
    if StreamDecompressor.check_format(fin):
        form = "framed"
    elif HadoopStreamDecompressor.check_format(fin):
        form = "hadoop"
    elif check_unframed_format(fin, reset=True):
        form = "raw"
    else:
        raise UncompressError("Can't detect format")
    return form, _DECOMPRESS_FORMAT_FUNCS[form]


def get_decompress_function(specified_format, fin):
    if specified_format == "auto":
        format, decompress_func = guess_format_by_header(fin)
        return decompress_func
    return _DECOMPRESS_METHODS[specified_format]


def get_compress_function(specified_format):
    if specified_format == "auto":
        return _COMPRESS_METHODS[_DEFAULT_COMPRESS_FORMAT]
    return _COMPRESS_METHODS[specified_format]



from inspect import isawaitable
from typing import TYPE_CHECKING
from typing import List
from uuid import uuid4

from .callbacks import CallbackGroup
from .event_data import TriggerData
from .exceptions import InvalidDefinition
from .i18n import _
from .transition_mixin import AddCallbacksMixin
from .utils import run_async_from_sync

if TYPE_CHECKING:
    from .statemachine import StateMachine
    from .transition_list import TransitionList


_event_data_kwargs = {
    "event_data",
    "machine",
    "event",
    "model",
    "transition",
    "state",
    "source",
    "target",
}


class Event(AddCallbacksMixin, str):
    """An event is triggers a signal that something has happened.

    They are send to a state machine and allow the state machine to react.

    An event starts a :ref:`Transition`, which can be thought of as a “cause” that initiates a
    change in the state of the system.

    See also :ref:`events`.
    """

    id: str
    """The event identifier."""

    name: str
    """The event name."""

    _sm: "StateMachine | None" = None
    """The state machine instance."""

    _transitions: "TransitionList | None" = None
    _has_real_id = False

    def __new__(
        cls,
        transitions: "str | TransitionList | None" = None,
        id: "str | None" = None,
        name: "str | None" = None,
        _sm: "StateMachine | None" = None,
    ):
        if isinstance(transitions, str):
            id = transitions
            transitions = None

        _has_real_id = id is not None
        id = str(id) if _has_real_id else f"__event__{uuid4().hex}"

        instance = super().__new__(cls, id)
        instance.id = id
        if name:
            instance.name = name
        elif _has_real_id:
            instance.name = str(id).replace("_", " ").capitalize()
        else:
            instance.name = ""
        if transitions:
            instance._transitions = transitions
        instance._has_real_id = _has_real_id
        instance._sm = _sm
        return instance

    def __repr__(self):
        return f"{type(self).__name__}({self.id!r})"

    def is_same_event(self, *_args, event: "str | None" = None, **_kwargs) -> bool:
        return self == event

    def _add_callback(self, callback, grouper: CallbackGroup, is_event=False, **kwargs):
        if self._transitions is None:
            raise InvalidDefinition(
                _("Cannot add callback '{}' to an event with no transitions.").format(callback)
            )
        return self._transitions._add_callback(
            callback=callback,
            grouper=grouper,
            is_event=is_event,
            **kwargs,
        )

    def __get__(self, instance, owner):
        """By implementing this method `Event` can be used as a property descriptor

        When attached to a SM class, if the user tries to get the `Event` instance,
        we intercept here and return a `BoundEvent` instance, so the user can call
        it as a method with the correct SM instance.

        """
        if instance is None:
            return self
        return BoundEvent(id=self.id, name=self.name, _sm=instance)

    def __call__(self, *args, **kwargs):
        """Send this event to the current state machine.

        Triggering an event on a state machine means invoking or sending a signal, initiating the
        process that may result in executing a transition.
        """
        # The `__call__` is declared here to help IDEs knowing that an `Event`
        # can be called as a method. But it is not meant to be called without
        # an SM instance. Such SM instance is provided by `__get__` method when
        # used as a property descriptor.
        machine = self._sm
        if machine is None:
            raise RuntimeError(_("Event {} cannot be called without a SM instance").format(self))

        kwargs = {k: v for k, v in kwargs.items() if k not in _event_data_kwargs}
        trigger_data = TriggerData(
            machine=machine,
            event=self,
            args=args,
            kwargs=kwargs,
        )
        machine._put_nonblocking(trigger_data)
        result = machine._processing_loop()
        if not isawaitable(result):
            return result
        return run_async_from_sync(result)

    def split(  # type: ignore[override]
        self, sep: "str | None" = None, maxsplit: int = -1
    ) -> List["Event"]:
        result = super().split(sep, maxsplit)
        if len(result) == 1:
            return [self]
        return [Event(event) for event in result]


class BoundEvent(Event):
    pass



from os import PathLike
from pathlib import Path
from typing import TypedDict, Callable, Optional, TypeVar, Any, Type, Union

import jinja2
from loguru import logger
import yaml
from celery import Task
from jinja2 import Environment, PrefixLoader, TemplateNotFound
from pydantic import BaseModel
from statemachine import State

from genie_flow.genie import GenieModel, GenieStateMachine
from genie_flow_invoker.pool import InvokersPool
from genie_flow_invoker.factory import InvokerFactory
from genie_flow.model.types import ModelKeyRegistryType
from genie_flow.model.template import CompositeTemplateType, MapTaskTemplate

_META_FILENAME: str = "meta.yaml"
_T = TypeVar("_T")


class _RegisteredDirectory(TypedDict):
    directory: Path
    jinja_loader: jinja2.FileSystemLoader
    config: dict
    invokers: Optional[InvokersPool]


class GenieEnvironment:
    """
    The `GenieEnvironment` deals with maintaining the templates registry, rendering templates
    and invoking `Invoker`s with a data context and a dialogue.
    """

    def __init__(
        self,
        template_root_path: str | PathLike,
        pool_size: int,
        model_key_registry: ModelKeyRegistryType,
        invoker_factory: InvokerFactory,
    ):
        self.template_root_path = Path(template_root_path).resolve()
        self.pool_size = pool_size
        self.model_key_registry = model_key_registry
        self.invoker_factory = invoker_factory

        self._jinja_env: Optional[Environment] = None
        self._template_directories: dict[str, _RegisteredDirectory] = {}

    def _walk_directory_tree_upward(
        self, start_directory: Path, execute: Callable[[Path, Optional[dict]], _T]
    ) -> _T:
        start_directory = start_directory.resolve()
        if start_directory == self.template_root_path:
            return execute(start_directory, None)

        parent_directory = start_directory.parent
        if parent_directory == start_directory:  # we reached the top-most directory
            raise ValueError("start_directory not part of the template directory tree")

        parent_result = self._walk_directory_tree_upward(parent_directory, execute)
        return execute(start_directory, parent_result)

    def _add_all_directories(self, start_directory: Path):
        start_directory = start_directory.resolve()
        for directory_element in start_directory.glob("*"):
            if directory_element.is_dir():
                self._add_all_directories(directory_element)
        self.register_template_directory(start_directory.name, start_directory)

    @staticmethod
    def read_meta(directory: Path, parent_config: Optional[dict]) -> dict:
        if parent_config is None:
            parent_config = {}
        try:
            with open(directory / _META_FILENAME, "r") as meta_file:
                meta = yaml.safe_load(meta_file)
                parent_config.update(meta)
                return parent_config
        except FileNotFoundError:
            logger.debug(f"No meta file found in {directory}")
            return parent_config

    @property
    def jinja_loader_mapping(self) -> dict[str, jinja2.BaseLoader]:
        return {
            prefix: directory["jinja_loader"]
            for prefix, directory in self._template_directories.items()
        }

    @property
    def jinja_env(self) -> jinja2.Environment:
        if self._jinja_env is None:
            self._jinja_env = Environment(
                loader=PrefixLoader(self.jinja_loader_mapping)
            )
        return self._jinja_env

    def _non_existing_templates(self, template: CompositeTemplateType) -> list[CompositeTemplateType]:
        if isinstance(template, str):
            try:
                _ = self.get_template(template)
                return []
            except TemplateNotFound:
                return [template]

        if isinstance(template, Task):
            # TODO might want to check if the task exists
            return []

        if isinstance(template, list):
            result = []
            for t in template:
                result.extend(self._non_existing_templates(t))
            return result

        if isinstance(template, dict):
            result = []
            for key in template.keys():
                result.extend(
                    [f"{key}:{t}" for t in self._non_existing_templates(template[key])]
                )
            return result

        if isinstance(template, MapTaskTemplate):
            return self._non_existing_templates(template.template_name)

        raise RuntimeError(f"Unknown template type: {type(template)}")

    def _validate_state_templates(self, state_machine_class: type[GenieStateMachine]):
        templates = state_machine_class.templates
        states_without_template = {
            state.id
            for state in state_machine_class.states
            if isinstance(state, State) and state.id not in templates
        }

        unknown_template_names = self._non_existing_templates(
            [
                templates[state.id]
                for state in state_machine_class.states
                if state not in states_without_template
            ]
        )

        if states_without_template or unknown_template_names:
            raise ValueError(
                f"GenieStateMachine {state_machine_class} is missing templates for states: ["
                f"{', '.join(states_without_template)}] and "
                f"cannot find templates with names: [{', '.join(unknown_template_names)}]"
            )

    def _validate_state_values(self, state_machine_class: type[GenieStateMachine]):
        state_values = [state.value for state in state_machine_class.states]
        state_values_set = set(state_values)
        duplicate_values = set()
        for state_value in state_values:
            try:
                state_values_set.remove(state_value)
            except KeyError:
                duplicate_values.add(state_value)

        if len(duplicate_values) > 0:
            raise ValueError(
                f"For GenieStateMachine {state_machine_class}, "
                f"the following values are duplicates: {duplicate_values}")

    def register_model(self, model_key: str, model_class: Type[BaseModel]):
        """
        Register a model class, so it can be stored in the object store. Also registers
        the model with the given model_key for the API.

        :param model_key: the key at which the genie flow is reachable for the given model_class
        :param model_class: the class of the model that needs to be registered
        """
        if not issubclass(model_class, GenieModel):
            raise ValueError(
                f"Can only register subclasses of GenieModel, not {model_class}"
            )

        self._validate_state_values(model_class.get_state_machine_class())
        self._validate_state_templates(model_class.get_state_machine_class())

        if model_key in self.model_key_registry:
            raise ValueError(f"Model key {model_key} already registered")

        self.model_key_registry[model_key] = model_class

    def _create_invoker_registration(
            self,
            directory_path: Path,
            invoker_config: dict,
    ) -> _RegisteredDirectory:
        nr_invokers = (
            self.pool_size
            if "pool_size" not in invoker_config
            else invoker_config["pool_size"]
        )
        return _RegisteredDirectory(
            directory=directory_path,
            config=invoker_config,
            jinja_loader=jinja2.FileSystemLoader(directory_path),
            invokers=self.invoker_factory.create_invoker_pool(
                nr_invokers,
                invoker_config,
            ),
        )

    def _create_renderer_registration(
            self,
            directory_path: Path,
            renderer_config: dict,
    ) -> _RegisteredDirectory:
        return _RegisteredDirectory(
            directory=directory_path,
            jinja_loader=jinja2.FileSystemLoader(directory_path),
            config=renderer_config,
            invokers=None,
        )

    def register_template_directory(self, prefix: str, directory: str | PathLike):
        if prefix in self._template_directories:
            raise ValueError(f"Template prefix '{prefix}' already registered")

        directory_path = Path(directory).resolve()
        config = self._walk_directory_tree_upward(directory_path, self.read_meta)

        if "invoker" in config and "renderer" in config:
            logger.error(
                "Compiled template directory {directory_path} configuration "
                "contains both 'invoker' and 'renderer'",
                directory_path=directory_path,
            )
            raise ValueError(
                f"Compiled template directory {directory_path} configuration "
                "contains both 'invoker' and 'renderer'"
            )

        if "invoker" in config:
            self._template_directories[prefix] = self._create_invoker_registration(
                directory_path,
                config["invoker"],
            )
        elif "renderer" in config:
            self._template_directories[prefix] = self._create_renderer_registration(
                directory_path,
                config["renderer"],
            )
        else:
            logger.error(
                "Compiled template directory {directory_path} configuration "
                "does not contain 'invoker' or 'renderer'",
                directory_path=directory_path,
            )
            raise ValueError(
                f"Compiled template directory {directory_path} configuration "
                "does not contain 'invoker' or 'renderer'"
            )

        self._jinja_env = None  # clear the Environment

    def has_invoker(self, template: CompositeTemplateType) -> bool:
        if not isinstance(template, str):
            return True

        prefix, _ = template.rsplit("/", 1)
        return self._template_directories[prefix]["invokers"] is not None

    def get_template(self, template_path: str) -> jinja2.Template:
        return self.jinja_env.get_template(template_path)

    def render_template(self, template_path: str, data_context: dict[str, Any]) -> str:
        template = self.jinja_env.get_template(template_path)
        rendered =  template.render(data_context)
        logger.debug(
            "rendered template {template_path} into {rendered}",
            template_path=template_path,
            rendered=rendered,
        )
        return rendered

    def invoke_template(
        self,
        template_path: str,
        data_context: dict[str, Any],
    ) -> str:
        rendered = self.render_template(template_path, data_context)

        prefix, _ = template_path.rsplit("/", 1)
        invokers_pool = self._template_directories[prefix]["invokers"]
        if invokers_pool is None:
            logger.error(
                "no invokers registered for template {template_path}",
                template_path=template_path,
            )
            raise ValueError(f"no invokers registered for template {template_path}")

        with invokers_pool as invoker:
            return invoker.invoke(rendered)



from typing import TYPE_CHECKING

from .i18n import _

if TYPE_CHECKING:
    from .event import Event
    from .state import State


class StateMachineError(Exception):
    "Base exception for this project, all exceptions that can be raised inherit from this class."


class InvalidDefinition(StateMachineError):
    "The state machine has a definition error"


class InvalidStateValue(InvalidDefinition):
    "The current model state value is not mapped to a state definition."

    def __init__(self, value, msg=None):
        self.value = value
        if msg is None:
            msg = _("{!r} is not a valid state value.").format(value)
        super().__init__(msg)


class AttrNotFound(InvalidDefinition):
    "There's no method or property with the given name"


class TransitionNotAllowed(StateMachineError):
    "Raised when there's no transition that can run from the current :ref:`state`."

    def __init__(self, event: "Event", state: "State"):
        self.event = event
        self.state = state
        msg = _("Can't {} when in {}.").format(self.event.name, self.state.name)
        super().__init__(msg)



from enum import Enum

from loguru import logger
from pydantic import BaseModel, Field, RootModel

from genie_flow.model.versioned import VersionedModel
from genie_flow.utils import get_fully_qualified_name_from_class, \
    get_class_from_fully_qualified_name


class PersistenceState(Enum):
    NEW_OBJECT = 0  # new object that should be persisted
    RETRIEVED_OBJECT = 1  # existing object that should NOT be persisted
    DELETED_OBJECT = 2  # old object that should be removed


class SecondaryStore(RootModel[dict[str, VersionedModel]]):
    """
    Represents a secondary data storage model that acts as a dict of str to `VersionedModel`
    instances. It tracks the state of its items for persistence purposes, ensuring that only
    values that are marked as NEW_OBJECT are stored.
    """
    root: dict[str, VersionedModel] = Field(default_factory=dict)
    _states: dict[str, Enum] = dict()

    def __getitem__(self, key: str) -> VersionedModel:
        return self.root[key]

    def __setitem__(self, key: str, value: VersionedModel):
        if not isinstance(value, VersionedModel):
            logger.error(
                "Cannot set a value for key {key} that is not a VersionedModel",
                key=key,
            )
            raise ValueError("Invalid data type for secondary store: must be a VersionedModel")

        self.root[key] = value
        self._states[key] = PersistenceState.NEW_OBJECT

    def __delitem__(self, key: str):
        del self.root[key]
        self._states[key] = PersistenceState.DELETED_OBJECT

    @classmethod
    def from_retrieved_values(cls, retrieved_values: dict[str, VersionedModel]) -> "SecondaryStore":
        """
        Create a SecondaryStore from retrieved values. This ensures that the state of
        all properties is set to RETRIEVED_OBJECT.

        :param retrieved_values: a dictionary that this SecondaryStore should encapsulate
        :return: a new SecondaryStore with the retrieved values as root values,
        and all states set to RETRIEVED_OBJECT
        """
        result = cls.model_validate(retrieved_values)
        for key in retrieved_values.keys():
            result._states[key] = PersistenceState.RETRIEVED_OBJECT
        return result

    @classmethod
    def from_serialized(cls, payloads: dict[str, bytes]) -> "SecondaryStore":
        """
        Create a SecondaryStore from serialized values. This ensures that the state of
        all properties states are set to RETRIEVED_OBJECT.

        :param payloads: a dictionary where the values for each key are serialized objects
        :return: a new SecondaryStore with the retrieved values as root values,
        and all states set to RETRIEVED_OBJECT
        """
        key_values: dict[str, VersionedModel] = dict()
        for key, payload in payloads.items():
            payload_type, payload = payload.split(b":", maxsplit=1)
            model_class = get_class_from_fully_qualified_name(payload_type.decode("utf-8"))
            if not issubclass(model_class, VersionedModel):
                logger.error(
                    "Cannot unserialize a payload with type {payload_type} that "
                    "is not a VersionedModel",
                    payload_type=payload_type,
                )
                raise ValueError(
                    f"Cannot unserialize a payload with type {payload_type} that "
                    f"is not a VersionedModel",
                )
            key_values[key] = model_class.deserialize(payload)
        return cls.from_retrieved_values(key_values)

    @property
    def has_unpersisted_values(self) -> bool:
        """Does this SecondaryStore have any unpersisted values?"""
        return any(
            state == PersistenceState.NEW_OBJECT
            for state in self._states.values()
        )

    @property
    def has_deleted_values(self) -> bool:
        """Does this SecondaryStore have any deleted values?"""
        return any(
            state == PersistenceState.DELETED_OBJECT
            for state in self._states.values()
        )

    @property
    def deleted_keys(self) -> set[str]:
        """Returns a set of keys that are marked as deleted"""
        return {
            key
            for key, state in self._states.items()
            if state == PersistenceState.DELETED_OBJECT
        }

    @property
    def unpersisted_values(self) -> dict[str, VersionedModel]:
        """Returns a dictionary of keys and values that are marked as new"""
        return {
            key: self.root[key]
            for key, state in self._states.items()
            if state == PersistenceState.NEW_OBJECT
        }

    def mark_persisted(self, keys: str | list[str]):
        """
        Marks the provided keys as persisted in the internal state dictionary. If any key is
        already marked as persisted, a KeyError is raised.

        :param keys: A string representing a single key or a list of strings representing
                     multiple keys to be marked as persisted.
        :raises KeyError: If attempting to mark a key as persisted when it is already marked
                          as such.
        """
        if isinstance(keys, str):
            keys = [keys]

        for key in keys:
            if self._states[key] == PersistenceState.RETRIEVED_OBJECT:
                logger.error(
                    "Trying to mark {key} as persisted, but it is already marked as such",
                    key=key,
                )
                raise KeyError("Attempting to overwrite existing persisted id")
            logger.debug("Marking {key} as persisted", key=key)
            self._states[key] = PersistenceState.RETRIEVED_OBJECT

    def unpersisted_serialized(self, compression: bool) -> dict[str, bytes]:
        """
        Serializes unpersisted values in the instance to a dictionary, encoding them with
        the proper format and optionally applying compression.

        The function iterates over the `unpersisted_values` attribute, retrieves the fully
        qualified name (FQN) of each value's class, serializes the value itself, and combines
        the FQN with the serialized data. It then stores the result in a dictionary where
        keys are the original keys, and values are the encoded serialized results.

        :param compression: Flag indicating whether to apply compression during serialization.
        :type compression: bool
        :return: A dictionary mapping the original keys to the encoded serialized data.
        :rtype: dict[str, bytes]
        """
        result: dict[str, bytes] = dict()
        for key, value in self.unpersisted_values.items():
            model_fqn = get_fully_qualified_name_from_class(value)
            value_serialized = value.serialize(compression)
            result[key] = model_fqn.encode("utf-8") + b":" + value_serialized
        return result



"""Secure serializer."""
from kombu.serialization import dumps, loads, registry
from kombu.utils.encoding import bytes_to_str, ensure_bytes, str_to_bytes

from celery.app.defaults import DEFAULT_SECURITY_DIGEST
from celery.utils.serialization import b64decode, b64encode

from .certificate import Certificate, FSCertStore
from .key import PrivateKey
from .utils import get_digest_algorithm, reraise_errors

__all__ = ('SecureSerializer', 'register_auth')

# Note: we guarantee that this value won't appear in the serialized data,
# so we can use it as a separator.
# If you change this value, make sure it's not present in the serialized data.
DEFAULT_SEPARATOR = str_to_bytes("\x00\x01")


class SecureSerializer:
    """Signed serializer."""

    def __init__(self, key=None, cert=None, cert_store=None,
                 digest=DEFAULT_SECURITY_DIGEST, serializer='json'):
        self._key = key
        self._cert = cert
        self._cert_store = cert_store
        self._digest = get_digest_algorithm(digest)
        self._serializer = serializer

    def serialize(self, data):
        """Serialize data structure into string."""
        assert self._key is not None
        assert self._cert is not None
        with reraise_errors('Unable to serialize: {0!r}', (Exception,)):
            content_type, content_encoding, body = dumps(
                data, serializer=self._serializer)

            # What we sign is the serialized body, not the body itself.
            # this way the receiver doesn't have to decode the contents
            # to verify the signature (and thus avoiding potential flaws
            # in the decoding step).
            body = ensure_bytes(body)
            return self._pack(body, content_type, content_encoding,
                              signature=self._key.sign(body, self._digest),
                              signer=self._cert.get_id())

    def deserialize(self, data):
        """Deserialize data structure from string."""
        assert self._cert_store is not None
        with reraise_errors('Unable to deserialize: {0!r}', (Exception,)):
            payload = self._unpack(data)
            signature, signer, body = (payload['signature'],
                                       payload['signer'],
                                       payload['body'])
            self._cert_store[signer].verify(body, signature, self._digest)
        return loads(body, payload['content_type'],
                     payload['content_encoding'], force=True)

    def _pack(self, body, content_type, content_encoding, signer, signature,
              sep=DEFAULT_SEPARATOR):
        fields = sep.join(
            ensure_bytes(s) for s in [b64encode(signer), b64encode(signature),
                                      content_type, content_encoding, body]
        )
        return b64encode(fields)

    def _unpack(self, payload, sep=DEFAULT_SEPARATOR):
        raw_payload = b64decode(ensure_bytes(payload))
        v = raw_payload.split(sep, maxsplit=4)
        return {
            'signer': b64decode(v[0]),
            'signature': b64decode(v[1]),
            'content_type': bytes_to_str(v[2]),
            'content_encoding': bytes_to_str(v[3]),
            'body': v[4],
        }


def register_auth(key=None, key_password=None, cert=None, store=None,
                  digest=DEFAULT_SECURITY_DIGEST,
                  serializer='json'):
    """Register security serializer."""
    s = SecureSerializer(key and PrivateKey(key, password=key_password),
                         cert and Certificate(cert),
                         store and FSCertStore(store),
                         digest, serializer=serializer)
    registry.register('auth', s.serialize, s.deserialize,
                      content_type='application/data',
                      content_encoding='utf-8')



from tests.legacy import unittest
from tests import test_lexer

from jmespath import lexer
from jmespath.visitor import Options
from jmespath.exceptions import LexerError


class TestLegacyRegexLexer(test_lexer.LexerUtils):

    def setUp(self):
        self.lexer = lexer.Lexer()
        self.options = Options(enable_legacy_literals=True)

    def tokenize(self, expression):
        return self.lexer.tokenize(expression, self.options)

    def test_literal_string(self):
        tokens = list(self.tokenize('`foobar`'))
        self.assert_tokens(tokens, [
            {'type': 'literal', 'value': "foobar"},
        ])

    def test_literal_with_invalid_json(self):
        with self.assertRaises(LexerError):
            list(self.tokenize('`foo"bar`'))

    def test_literal_with_empty_string(self):
        tokens = list(self.tokenize('``'))
        self.assert_tokens(tokens, [{'type': 'literal', 'value': ''}])

    def test_adds_quotes_when_invalid_json(self):
        tokens = list(self.tokenize('`{{}`'))
        self.assertEqual(
            tokens,
            [{'type': 'literal', 'value': '{{}',
              'start': 0, 'end': 4},
             {'type': 'eof', 'value': '',
              'start': 5, 'end': 5}
             ]
        )

if __name__ == '__main__':
    unittest.main()
